#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

"""
Initializer for the fixed bed TSA unit model.
"""

# Import Pyomo libraries
from pyomo.environ import (
    value,
    check_optimal_termination,
    Block,
    units,
    Var,
    Constraint,
)
from pyomo.util.calc_var_value import calculate_variable_from_constraint

# import IDAES core libraries
from idaes.core.initialization import ModularInitializerBase
from idaes.core.initialization.initializer_base import StoreState
from idaes.core.util.exceptions import InitializationError
from idaes.core.util.model_serializer import to_json, from_json
from idaes.core.util.model_statistics import degrees_of_freedom

from idaes.models_extra.temperature_swing_adsorption import SteamCalculationType


import idaes.logger as idaeslog

__author__ = "Daison Yancy Caballero, Alex Noring"

# Set up logger
_log = idaeslog.getLogger(__name__)


class FixedBedTSA0DInitializer(ModularInitializerBase):
    """
    Initializer for 0D Fixed Bed TSA units.

    """

    # TODO: Using IPOPT_v2 as the default solver causes lots of test failures
    CONFIG = ModularInitializerBase.CONFIG()
    CONFIG.solver = "ipopt"

    def initialize(
        self,
        model: Block,
        initial_guesses: dict = None,
        json_file: str = None,
        output_level=None,
        exclude_unused_vars: bool = True,
        heating_time_guess=1000,
        cooling_time_guess=500,
    ):

        if not exclude_unused_vars:
            exclude_unused_vars = True
            _log.warning(
                "The FixedBedTSA0DInitializer was run with "
                "exclude_unused_vars=False, but the FixedBedTSA0D model "
                "contains unused Vars. exclude_unused_vars has been set "
                "to True to avoid raising an InitializationError."
            )

        self.heating_time_guess = heating_time_guess
        self.cooling_time_guess = cooling_time_guess

        super().initialize(
            model=model,
            initial_guesses=initial_guesses,
            json_file=json_file,
            output_level=output_level,
            exclude_unused_vars=exclude_unused_vars,
        )

    def initialization_routine(
        self,
        blk: Block,
    ):
        """
        Initialization routine for fixed bed TSA unit.

        Args:
            blk: model to be initialized

        Returns:
            None

        Raises:
            InitializationError: If degrees of freedom is not zero at the start
            of each initialization step.

        """
        # set up logger for initialization and solve
        init_log = idaeslog.getInitLogger(blk.name, self.get_output_level(), tag="unit")
        solve_log = idaeslog.getSolveLogger(
            blk.name, self.get_output_level(), tag="unit"
        )

        # create solver
        solver = self._get_solver()

        # initialization of fixed bed TSA model unit
        init_log.info("Starting fixed bed TSA initialization")

        tsa_state = to_json(blk, wts=StoreState, return_dict=True)

        # 1 - solve heating step

        # 1.1) fix states in the fixed bed TSA inlet  ("flow_mol_in_total",
        # "mol  e_frac_in", and "pressure_adsorption"). These are equal
        # to those states coming from the exhaust gas stream in the CCS system

        vars_lst_heating = ["flow_mol_in_total", "pressure_adsorption", "mole_frac_in"]

        cons_lst_heating = ["flow_mol_in_total_eq", "pressure_in_eq", "mole_frac_in_eq"]

        self._calculate_and_fix_variable_from_constraint(
            blk, variable_list=vars_lst_heating, constraint_list=cons_lst_heating
        )

        # 1.2) initial solution using false position method

        # deactivate final condition constraint and fix time
        blk.heating.fc_temperature_eq.deactivate()
        blk.heating.time.fix(self.heating_time_guess * units.s)

        # check degrees of freedom and solve
        if degrees_of_freedom(blk.heating) == 0:
            self._false_position_method(
                blk,
                cycle_step=blk.heating,
                t_guess=self.heating_time_guess,
            )
        else:
            raise InitializationError(
                "Degrees of freedom is not zero during initialization of "
                "heating step. Fix/unfix appropriate number of variables "
                "to result in zero degrees of freedom for initialization."
            )

        # 1.3) activate final condition constraint and solve entire step

        blk.heating.fc_temperature_eq.activate()
        blk.heating.time.unfix()

        # check degrees of freedom and solve
        if degrees_of_freedom(blk.heating) == 0:
            self._step_initialize(cycle_step=blk.heating)
        else:
            raise InitializationError(
                "Degrees of freedom is not zero during initialization of "
                "heating step. Fix/unfix appropriate number of variables "
                "to result in zero degrees of freedom for initialization."
            )

        # 2 - solve cooling step

        # 2.1) fix mole fraction at end of heating step

        vars_lst_cooling = ["mole_frac_heating_end"]
        cons_lst_cooling = ["mole_frac_heating_end_eq"]

        self._calculate_and_fix_variable_from_constraint(
            blk, variable_list=vars_lst_cooling, constraint_list=cons_lst_cooling
        )

        # 2.2) initial solution using false position method

        # deactivate final condition constraint and fix time
        blk.cooling.fc_temperature_eq.deactivate()
        blk.cooling.time.fix(self.cooling_time_guess * units.s)

        # check degrees of freedom and solve
        if degrees_of_freedom(blk.cooling) == 0:
            self._false_position_method(
                blk,
                cycle_step=blk.cooling,
                t_guess=self.cooling_time_guess,
            )
        else:
            raise InitializationError(
                "Degrees of freedom is not zero during initialization of "
                "cooling step. Fix/unfix appropriate number of variables "
                "to result in zero degrees of freedom for initialization."
            )

        # 2.3) activate final condition constraint and solve entire model

        blk.cooling.fc_temperature_eq.activate()
        blk.cooling.time.unfix()

        # check degrees of freedom and solve
        if degrees_of_freedom(blk.cooling) == 0:
            self._step_initialize(cycle_step=blk.cooling)
        else:
            raise InitializationError(
                "Degrees of freedom is not zero during initialization of "
                "cooling step. Fix/unfix appropriate number of variables "
                "to result in zero degrees of freedom for initialization."
            )

        # 3 - solve pressurization step

        # 3.1) fix mole fraction, temperature, pressure and loadings at
        # end of cooling step

        vars_lst_pressurization = [
            "mole_frac_cooling_end",
            "pressure_cooling_end",
            "loading_cooling_end",
        ]

        cons_lst_pressurization = [
            "mole_frac_cooling_end_eq",
            "pressure_cooling_end_eq",
            "loading_cooling_end_eq",
        ]

        self._calculate_and_fix_variable_from_constraint(
            blk,
            variable_list=vars_lst_pressurization,
            constraint_list=cons_lst_pressurization,
        )

        if blk.calculate_beds:
            # if "calculate_beds" is True, there is an extra variable for
            # "velocity_in" and it needs to be fixed to initialize the
            # pressurization and adsorption step
            velocity_fixed = blk.velocity_in.fixed
            if not velocity_fixed:
                blk.velocity_in.fix()
                blk.pressure_drop.unfix()

        # 3.2) check degrees of freedom and solve

        if degrees_of_freedom(blk.pressurization) == 0:
            self._step_initialize(
                cycle_step=blk.pressurization,
            )
        else:
            raise InitializationError(
                "Degrees of freedom is not zero during initialization of "
                "pressurization step. Fix/unfix appropriate number of "
                "variables to result in zero degrees of freedom for "
                "initialization."
            )

        # 4 - solve adsorption step

        # 4.1) fix mole fraction and loadings at end of pressurization step

        vars_lst_adsorption = [
            "mole_frac_pressurization_end",
            "loading_pressurization_end",
        ]

        cons_lst_adsorption = [
            "mole_frac_pressurization_end_eq",
            "loading_pressurization_end_eq",
        ]

        self._calculate_and_fix_variable_from_constraint(
            blk, variable_list=vars_lst_adsorption, constraint_list=cons_lst_adsorption
        )

        # 4.2) check degrees of freedom and solve

        if degrees_of_freedom(blk.adsorption) == 0:
            self._step_initialize(cycle_step=blk.adsorption)
        else:
            raise InitializationError(
                "Degrees of freedom is not zero during initialization of "
                "adsorption step. Fix/unfix appropriate number of variables "
                "to result in zero degrees of freedom for initialization."
            )

        # 5 - solve entire fixed bed TSA model

        # 5.1) unfix variables and activate constraints that were fixed and
        # deactivated during individual steps
        from_json(blk, sd=tsa_state, wts=StoreState)

        if blk.calculate_beds:
            calculate_variable_from_constraint(blk.velocity_in, blk.pressure_drop_eq)
        else:
            calculate_variable_from_constraint(blk.pressure_drop, blk.pressure_drop_eq)

        # 5.2) deactivate compressor
        if blk.config.compressor:
            blk.compressor.deactivate()

        # 5.3) deactivate steam calculation constraints
        if blk.config.steam_calculation != SteamCalculationType.none:
            if blk.config.steam_calculation == SteamCalculationType.rigorous:
                blk.steam_heater.deactivate()
            blk.flow_mass_steam_eq.deactivate()
            blk.flow_mass_steam.fix()

        # 5.4) check degrees of freedom and solve
        if degrees_of_freedom(blk) == 0:

            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solver.solve(blk, tee=slc.tee)

            if check_optimal_termination(res):
                init_log.info(
                    "Initialization of fixed bed TSA model "
                    "completed {}.".format(idaeslog.condition(res))
                )
            else:
                _log.warning(
                    "Initialization of fixed bed TSA model "
                    "Failed {}.".format(blk.name)
                )
        else:
            raise InitializationError(
                "Degrees of freedom is not zero during initialization of "
                "fixed bed TSA model. Fix/unfix appropriate number of "
                "variables to result in zero degrees of freedom for "
                "initialization."
            )

        # 6 - solve compressor unit
        if blk.config.compressor:

            # initialization of compressor
            init_log.info("Starting initialization of compressor.")

            # activate compressor model
            blk.compressor.activate()

            # 6.1) fix state at inlet of compressor to match states in
            # exhaust gas stream in the CCS system. Fix pressure drop in
            # compressor unit.
            for t in blk.flowsheet().time:
                for i in blk.config.compressor_properties.component_list:
                    calculate_variable_from_constraint(
                        blk.compressor.unit.inlet.flow_mol_comp[t, i],
                        blk.compressor.flow_mol_in_compressor_eq[t, i],
                    )
                calculate_variable_from_constraint(
                    blk.compressor.unit.inlet.temperature[t],
                    blk.compressor.temperature_in_compressor_eq[t],
                )
                calculate_variable_from_constraint(
                    blk.compressor.unit.inlet.pressure[t],
                    blk.compressor.pressure_in_compressor_eq[t],
                )
                calculate_variable_from_constraint(
                    blk.compressor.unit.deltaP[t],
                    blk.compressor.pressure_drop_tsa_compressor_eqn[t],
                )

            blk.compressor.unit.inlet.flow_mol_comp[:, :].fix()
            blk.compressor.unit.inlet.temperature[:].fix()
            blk.compressor.unit.inlet.pressure[:].fix()
            blk.compressor.unit.deltaP[:].fix()

            # deactivate related constraints for fixed variables.
            blk.compressor.flow_mol_in_compressor_eq.deactivate()
            blk.compressor.temperature_in_compressor_eq.deactivate()
            blk.compressor.pressure_in_compressor_eq.deactivate()
            blk.compressor.pressure_drop_tsa_compressor_eqn.deactivate()

            # 6.2) check degrees of freedom and solve
            if degrees_of_freedom(blk.compressor) == 0:

                # TODO: switch to new initialization method when implemented
                # for FlueGasStateBlock
                blk.compressor.unit.initialize()

                # re-solve compressor model
                with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                    res = solver.solve(blk.compressor, tee=slc.tee)

                if check_optimal_termination(res):
                    init_log.info(
                        "Initialization of compressor completed {}.".format(
                            idaeslog.condition(res)
                        )
                    )
                else:
                    _log.warning(
                        "Initialization of compressor Failed {}.".format(
                            blk.compressor.unit.name
                        )
                    )
            else:
                raise InitializationError(
                    "Degrees of freedom is not zero during initialization of "
                    "compressor model. Fix/unfix appropriate number of "
                    "variables to result in zero degrees of freedom for "
                    "initialization."
                )

        # 7 - solve steam calculation
        if blk.config.steam_calculation != SteamCalculationType.none:

            if blk.config.steam_calculation == SteamCalculationType.rigorous:

                # initialization of steam heater
                init_log.info(
                    "Starting initialization of heater model for steam calculation."
                )

                # activate steam heater model
                blk.steam_heater.activate()

                # 7.1) deactivate constraints for total saturation condition
                #      and heat_duty_heater_eq
                blk.steam_heater.unit.vapor_frac_out_eq.deactivate()
                blk.steam_heater.unit.heat_duty_heater_eq.deactivate()

                # 7.2) assume a dumy inlet flow rate and heat duty and fix them
                Fin = 100 * units.mol / units.s
                Q = -500000 * units.W
                blk.steam_heater.unit.inlet.flow_mol[0].fix(Fin)
                blk.steam_heater.unit.heat_duty.fix(Q)

                # 7.3) check degrees of freedom and solve
                if degrees_of_freedom(blk.steam_heater) == 0:

                    # initialize steam heater
                    heater_initializer = self.get_submodel_initializer(
                        blk.steam_heater.unit
                    )
                    heater_initializer.initialize(blk.steam_heater.unit)

                else:
                    raise InitializationError(
                        "Degrees of freedom is not zero during initialization "
                        "of heater model. Fix/unfix appropriate number of "
                        "variables to result in zero degrees of freedom for "
                        "initialization."
                    )

                # 7.4) activate constraint for total saturation condition and
                #      unfix heat duty
                blk.steam_heater.unit.vapor_frac_out_eq.activate()
                blk.steam_heater.unit.heat_duty.unfix()

                # 7.5) solve model for total saturation at outlet
                if degrees_of_freedom(blk.steam_heater) == 0:

                    init_log.info_high(
                        "Starting initialization of heater model "
                        "for total saturation at outlet."
                    )

                    with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                        res = solver.solve(blk.steam_heater, tee=slc.tee)

                    if check_optimal_termination(res):
                        init_log.info_high(
                            "Initialization of heater model "
                            "for total saturation at outlet "
                            "completed {}.".format(idaeslog.condition(res))
                        )
                    else:
                        _log.warning(
                            "Initialization of heater model for "
                            "total saturation at outlet Failed {}.".format(
                                blk.steam_heater.unit.name
                            )
                        )
                else:
                    raise InitializationError(
                        "Degrees of freedom is not zero during initialization "
                        "of heater model for total saturation at outlet. "
                        "Fix/unfix appropriate number of variables to result "
                        "in zero degrees of freedom for initialization."
                    )

                # 7.6) unfix inlet flow rate and fix heat duty
                blk.steam_heater.unit.inlet.flow_mol[0].unfix()
                calculate_variable_from_constraint(
                    blk.steam_heater.unit.heat_duty[0],
                    blk.steam_heater.unit.heat_duty_heater_eq,
                )
                blk.steam_heater.unit.heat_duty.fix()

                # 7.7) solve model for steam flow rate
                if degrees_of_freedom(blk.steam_heater) == 0:

                    init_log.info_high(
                        "Starting initialization of heater model "
                        "for total steam flow rate."
                    )

                    with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                        res = solver.solve(blk.steam_heater, tee=slc.tee)

                    if check_optimal_termination(res):
                        init_log.info_high(
                            "Initialization of heater model "
                            "for total steam flow rate "
                            "completed: {}.".format(idaeslog.condition(res))
                        )
                        init_log.info(
                            "Initialization of heater model "
                            "for steam calculation completed {}.".format(
                                idaeslog.condition(res)
                            )
                        )
                    else:
                        _log.warning(
                            "Initialization of heater model for "
                            "total steam flow rate Failed {}.".format(
                                blk.steam_heater.unit.name
                            )
                        )
                else:
                    raise InitializationError(
                        "Degrees of freedom is not zero during initialization "
                        "of heater model for total steam flow rate. "
                        "Fix/unfix appropriate number of variables to result "
                        "in zero degrees of freedom for initialization."
                    )

            calculate_variable_from_constraint(
                blk.flow_mass_steam, blk.flow_mass_steam_eq
            )

        # 8 - solve fixed bed TSA model, steam calculation constraints and
        #     compressor unit simultaneously

        if blk.config.compressor:

            # 8.1) unfix state that were fixed in 6.1
            blk.compressor.unit.inlet.flow_mol_comp[:, :].unfix()
            blk.compressor.unit.inlet.temperature[:].unfix()
            blk.compressor.unit.inlet.pressure[:].unfix()
            blk.compressor.unit.deltaP[:].unfix()

            # activate constraints that were deactivated in 6.1
            blk.compressor.flow_mol_in_compressor_eq.activate()
            blk.compressor.temperature_in_compressor_eq.activate()
            blk.compressor.pressure_in_compressor_eq.activate()
            blk.compressor.pressure_drop_tsa_compressor_eqn.activate()

        if blk.config.steam_calculation != SteamCalculationType.none:

            # 8.2 unfix variables and activate constraints that were fixed
            #     and deactivated in 7
            if blk.config.steam_calculation == SteamCalculationType.rigorous:
                blk.steam_heater.unit.heat_duty_heater_eq.activate()
                blk.steam_heater.unit.heat_duty.unfix()

            blk.flow_mass_steam_eq.activate()
            blk.flow_mass_steam.unfix()

        # 8.3) check degrees of freedom and solve
        if (
            blk.config.compressor
            or blk.config.steam_calculation != SteamCalculationType.none
        ):
            if degrees_of_freedom(blk) == 0:

                with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                    res = solver.solve(blk, tee=slc.tee)
                if (
                    blk.config.compressor
                    and blk.config.steam_calculation != SteamCalculationType.none
                ):
                    if check_optimal_termination(res):
                        init_log.info(
                            "Initialization of fixed bed TSA, steam "
                            "calculation and compressor models completed {}"
                            ".".format(idaeslog.condition(res))
                        )
                    else:
                        _log.warning(
                            "Initialization of fixed bed TSA, "
                            "steam calculation and compressor "
                            "models Failed {}.".format(blk.name)
                        )
                if (
                    blk.config.compressor
                    and blk.config.steam_calculation != SteamCalculationType.none
                ):
                    if check_optimal_termination(res):
                        init_log.info(
                            "Initialization of fixed bed TSA "
                            "and compressor models completed {}.".format(
                                idaeslog.condition(res)
                            )
                        )
                    else:
                        _log.warning(
                            "Initialization of fixed bed TSA "
                            "and compressor models Failed {}.".format(blk.name)
                        )
                if (
                    not blk.config.compressor
                    and blk.config.steam_calculation != SteamCalculationType.none
                ):
                    if check_optimal_termination(res):
                        init_log.info(
                            "Initialization of fixed bed TSA and steam "
                            "calculation models completed {}"
                            ".".format(idaeslog.condition(res))
                        )
                    else:
                        _log.warning(
                            "Initialization of fixed bed TSA "
                            "and steam calculation "
                            "models Failed {}.".format(blk.name)
                        )
            else:
                if (
                    blk.config.compressor
                    and blk.config.steam_calculation != SteamCalculationType.none
                ):
                    raise InitializationError(
                        "Degrees of freedom is not zero during initialization "
                        "of fixed bed TSA, steam calculation and compressor "
                        "models. Fix/unfix appropriate number of variables to "
                        "result in zero degrees of freedom for "
                        "initialization."
                    )
                if (
                    blk.config.compressor
                    and blk.config.steam_calculation == SteamCalculationType.none
                ):
                    raise InitializationError(
                        "Degrees of freedom is not zero during initialization "
                        "of fixed bed TSA and compressor "
                        "models. Fix/unfix appropriate number of variables to "
                        "result in zero degrees of freedom for "
                        "initialization."
                    )
                if (
                    not blk.config.compressor
                    and blk.config.steam_calculation != SteamCalculationType.none
                ):
                    raise InitializationError(
                        "Degrees of freedom is not zero during initialization "
                        "of fixed bed TSA and steam calculation "
                        "models. Fix/unfix appropriate number of variables to "
                        "result in zero degrees of freedom for "
                        "initialization."
                    )

    def _step_initialize(self, cycle_step=None):
        """
        Initialization routine for TSA cycle steps.

        Keyword Arguments:
            outlvl    : output level of initialisation routine
            solver    : str indicating which solver to use during
                        initialization
            optarg    : dictionary with solver options
            cycle_step: block model for cycle step

        """
        # set up logger
        init_log = idaeslog.getInitLogger(
            cycle_step.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            cycle_step.name, self.get_output_level(), tag="unit"
        )

        # initialization of cycle steps
        init_log.info(
            "Starting initialization of "
            + str(cycle_step).rsplit(".", maxsplit=1)[-1]
            + " step."
        )

        # solve cycle step
        solver = self._get_solver()
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(cycle_step, tee=slc.tee)

        if check_optimal_termination(res):
            init_log.info(
                "Initialization of "
                + str(cycle_step).rsplit(".", maxsplit=1)[-1]
                + " step completed {}.".format(idaeslog.condition(res))
            )
        else:
            _log.warning(
                "Initialization of "
                + str(cycle_step).rsplit(".", maxsplit=1)[-1]
                + " step Failed {}.".format(cycle_step.name)
            )

    def _false_position_method(self, blk, cycle_step=None, t_guess=None):
        """
        False position method to provide initial solution for TSA cycle steps.

        Keyword Arguments:
            outlvl    : output level of initialisation routine
            solver    : str indicating which solver to use during
                        initialization
            optarg    : dictionary with solver options
            cycle_step: block model for cycle step
            x0        : initial guess for time

        """
        # set up logger
        init_log = idaeslog.getInitLogger(
            cycle_step.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            cycle_step.name, self.get_output_level(), tag="unit"
        )

        # initial interval containing a root to apply false position method
        init_log.info_high(
            "Initialization of "
            + str(cycle_step).rsplit(".", maxsplit=1)[-1]
            + " step: step 1a: finding initial interval containing a root"
            " to apply false position method"
        )

        # fix time to initial guess
        x0 = t_guess
        cycle_step.time.fix(x0)

        # counter
        count = 1

        # solve model
        solver = self._get_solver()
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(cycle_step, tee=slc.tee)

        if check_optimal_termination(res):
            init_log.info_high(
                "Initialization of "
                + str(cycle_step).rsplit(".", maxsplit=1)[-1]
                + " step: step 1a - iteration {0}, completed {1}.".format(
                    count, idaeslog.condition(res)
                )
            )
        else:
            _log.warning(
                "Initialization of "
                + str(cycle_step).rsplit(".", maxsplit=1)[-1]
                + " step: step 1a - iteration {0}, Failed {1}.".format(
                    count, cycle_step.name
                )
            )

        # save solution in initial guess, f(x0)
        if str(cycle_step).rsplit(".", maxsplit=1)[-1] == "heating":
            f_x0 = value(blk.temperature_desorption - cycle_step.temperature[1])
        else:
            f_x0 = value(cycle_step.temperature[1] - blk.temperature_adsorption)
        f_x0_start = f_x0

        # iterate to find an interval with the root
        condition = True
        while condition:
            # update counter and x0
            count += 1

            if f_x0_start >= 0:
                x_new = 1.2 * x0
            else:
                x_new = x0 / 2.0

            # fix time to x_new guess
            cycle_step.time.fix(x_new)

            # solve model
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solver.solve(cycle_step, tee=slc.tee)

            if check_optimal_termination(res):
                init_log.info_high(
                    "Initialization of "
                    + str(cycle_step).rsplit(".", maxsplit=1)[-1]
                    + " step: step 1a - iteration {0}, completed {1}.".format(
                        count, idaeslog.condition(res)
                    )
                )
            else:
                _log.warning(
                    "Initialization of "
                    + str(cycle_step).rsplit(".", maxsplit=1)[-1]
                    + " step: step 1a - iteration {0}, Failed {1}.".format(
                        count, cycle_step.name
                    )
                )

            # save solution in new initial guess, f(x_new)
            if str(cycle_step).rsplit(".", maxsplit=1)[-1] == "heating":
                f_x_new = value(blk.temperature_desorption - cycle_step.temperature[1])
            else:
                f_x_new = value(cycle_step.temperature[1] - blk.temperature_adsorption)

            # set up new interval until find one containing the root
            if f_x0_start >= 0:
                if f_x_new >= 0:
                    x0 = x_new
                    f_x0 = f_x_new
                else:
                    x1 = x_new
                    f_x1 = f_x_new
                    condition = False
            else:
                if f_x_new >= 0:
                    x1 = x_new
                    f_x1 = f_x_new
                    condition = False
                else:
                    x0 = x_new
                    f_x0 = f_x_new

        # implementing false position method
        init_log.info_high(
            "Initialization of "
            + str(cycle_step).rsplit(".", maxsplit=1)[-1]
            + " step: step 1b: implementing false position method"
        )

        # counter
        count = 1
        condition = True

        # check condition to stop
        while condition:

            # compute new approximated root as x2
            x2 = x0 - (x1 - x0) * f_x0 / (f_x1 - f_x0)

            # fix time to x_2 guess
            cycle_step.time.fix(x2)

            # solve model
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solver.solve(cycle_step, tee=slc.tee)

            if check_optimal_termination(res):
                init_log.info_high(
                    "Initialization of "
                    + str(cycle_step).rsplit(".", maxsplit=1)[-1]
                    + " step: step 1b - iteration {0}, completed {1}.".format(
                        count, idaeslog.condition(res)
                    )
                )
            else:
                _log.warning(
                    "Initialization of "
                    + str(cycle_step).rsplit(".", maxsplit=1)[-1]
                    + " step: step 1b - iteration {0}, Failed {1}.".format(
                        count, cycle_step.name
                    )
                )

            # save solution in new x_2, f(x_2)
            if str(cycle_step).rsplit(".", maxsplit=1)[-1] == "heating":
                f_x2 = value(blk.temperature_desorption - cycle_step.temperature[1])
            else:
                f_x2 = value(cycle_step.temperature[1] - blk.temperature_adsorption)

            # check if f(x_0)*f(x_2) is negative
            if f_x0 * f_x2 < 0:
                x1 = x2
            else:
                x0 = x2

            # update counter and set up new condition |f(x_2)| > error
            count += 1
            condition = (f_x2**2) ** 0.5 > 1

    def _calculate_and_fix_variable_from_constraint(
        self, obj, variable_list=None, constraint_list=None
    ):
        """
        Method to calculate variables from a constraints, then fix the
        variables and deactivate the constraints.

        Keyword Arguments:
            variable_list: a list of str with names of variables to be
                calculated.
            constraint_list: a list of str with names of constraints
                used to calculate the variables in variable_list.

        """
        if variable_list is None:
            variable_list = []
        if constraint_list is None:
            constraint_list = []

        v_list = []
        c_list = []

        for v in obj.component_objects(Var, descend_into=True):
            if v.local_name in variable_list:
                v_list.append(v)
        for c in obj.component_objects(Constraint, descend_into=True):
            if c.local_name in constraint_list:
                c_list.append(c)

        v_c_tuple = list(zip(v_list, c_list))

        for k in v_c_tuple:
            if k[1].dim() == 0:
                calculate_variable_from_constraint(k[0], k[1])
            elif k[1].dim() == 1:
                for var_index in k[1].index_set():
                    calculate_variable_from_constraint(k[0][var_index], k[1][var_index])

        for v in v_list:
            v.fix()

        for c in c_list:
            c.deactivate()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Transformation to replace variables with other variables."""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.core.base.transformation import TransformationFactory
from pyomo.core.plugins.transform.hierarchy import NonIsomorphicTransformation
from pyomo.core import expr as EXPR
from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    document_kwargs_from_configdict,
)
from pyomo.core.base.var import VarData, Var
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.expression import Expression
from pyomo.core.base.objective import Objective


__author__ = "John Eslick"


def _is_var(v):
    return isinstance(v, (VarData, Var))


@TransformationFactory.register(
    "replace_variables", doc="Replace variables with other variables."
)
@document_kwargs_from_configdict("CONFIG")
class ReplaceVariables(NonIsomorphicTransformation):
    """Replace variables in a model or block with other variables.

    Keyword arguments below are specified for the ``apply_to(instance, **kwargs)``
    method.

    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "substitute",
        ConfigValue(
            default=[],
            description="List-like of tuples where the first item in a tuple is a "
            "Pyomo variable to be replaced and the second item in the "
            "tuple is a Pyomo variable to replace it with. This "
            "transformation is not reversible.",
        ),
    )

    @staticmethod
    def replace(instance, substitute):
        # Create the replacement dict. Do some argument validation and indexed
        # var handling
        d = {}
        for r in substitute:
            if not (_is_var(r[0]) and _is_var(r[1])):
                raise TypeError(
                    "Replace only allows variables to be replaced, {} is type {}"
                    " and {} is type {}".format(r[0], type(r[0]), r[1], type(r[1]))
                )
            if r[0].is_indexed() != r[1].is_indexed():
                raise TypeError(
                    "IndexedVars must be replaced by IndexedVars, {} is type {}"
                    " and {} is type {}".format(r[0], type(r[0]), r[1], type(r[1]))
                )
            if r[0].is_indexed() and r[1].is_indexed():
                if not r[0].index_set().issubset(r[1].index_set()):
                    raise ValueError(
                        "The index set of {} must be a subset of"
                        " {}.".format(r[0], r[1])
                    )
                for i in r[0]:
                    d[id(r[0][i])] = r[1][i]
            else:
                # scalar replace
                d[id(r[0])] = r[1]

        # Replacement Visitor
        vis = EXPR.ExpressionReplacementVisitor(
            substitute=d,
            descend_into_named_expressions=True,
            remove_named_expressions=False,
        )

        # Do replacements in Expressions, Constraints, and Objectives
        for c in instance.component_data_objects(
            (Constraint, Expression, Objective), descend_into=True, active=True
        ):
            c.set_value(expr=vis.walk_expression(c.expr))

    def _apply_to(self, instance, **kwds):
        """
        Apply the transformation.  This is called by ``apply_to`` in the
        superclass, and should not be called directly.  ``apply_to`` takes the
        same arguments.

        Args:
            instance: A block or model to apply the transformation to
            substitute: A list-like of two-element list-likes.  Each two element
                list-like specifies a replacement of the first variable by the
                second.  SimpleVar, IndexedVar, VarData, and Reference are
                all accepted types.

        Returns:
            None
        """
        config = self.CONFIG(kwds)
        self.replace(instance, config.substitute)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This model contains equations relating to the electrochemistry occurring at the
triple phase boundary. The (half-cell) Nernst potential and activation
overpotential are calculated, resistance heating is added to the heat flux, and
the reaction rate is linked to the current density using the stoichiometric
coefficient provided for ``"e^-"``.

Boundary variables:
    - ``temperature_deviation[t, iz]``
    - ``heat_flux_x0[t, iz]``
    - ``heat_flux_x1[t, iz]``
    - ``conc_mol_comp_deviation[t, iz]``
    - ``material_flux[t, iz]``

Instances of ``Var`` that must be fixed:
    - ``activation_potential_alpha1``
    - ``activation_potential_alpha2``
    - ``exchange_current_exponent_comp[j_react]``: Power law exponent of
      component partial pressure for reacting gases. Set to stoichiometric
      coefficients if unknown.
    - ``exchange_current_log_preexponential_factor``: Logarithm of
      preexponential factor of exchange current density in amps per m**2
    - ``exchange_current_activation_energy``: Activation energy in expression
      giving exchange current density.

These are all parameters in the Butler-Volmer equation.
See Noren and Hoffman (2005) for a good introduction.

Noren, D. A., Hoffman, M. A. (2005). Clarifying the Butler-Volmer equation
and related approximations for calculating activation losses in solid oxide
fuel cell models. Journal of Power Sources, 152(1–2), 175–181.
https://doi.org/10.1016/j.jpowsour.2005.03.174
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

__author__ = "Douglas Allan"

import copy

from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
import pyomo.environ as pyo


from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.util.constants import Constants
import idaes.models_extra.power_generation.unit_models.soc_submodels.common as common
from idaes.models_extra.power_generation.unit_models.soc_submodels.common import (
    _gas_species_list,
    _all_species_list,
)
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.solvers import get_solver

import idaes.logger as idaeslog


@declare_process_block_class("SocTriplePhaseBoundary")
class SocTriplePhaseBoundaryData(UnitModelBlockData):
    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag",
            doc="No capacities or holdups, so no internal dynamics",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(domain=In([False]), default=False),
    )
    CONFIG.declare(
        "component_list",
        ConfigValue(
            domain=common._SubsetOf(_gas_species_list), description="List of components"
        ),
    )
    CONFIG.declare(
        "reaction_stoichiometry",
        ConfigValue(
            common._SubsetOf(_all_species_list),
            description="Stochiometric coefficients for component reactions on the triple phase boundary. Must contain "
            "term for number of electrons consumed/liberated.",
        ),
    )
    CONFIG.declare(
        "inert_species",
        ConfigValue(
            default=None,
            domain=common._SubsetOf(_gas_species_list),
            description="List of species that do not participate in "
            "reactions at the triple phase boundary.",
        ),
    )
    CONFIG.declare(
        "conc_mol_comp_ref",
        ConfigValue(
            default=None,
            description="Variable for the component concentration in bulk channel ",
        ),
    )
    CONFIG.declare(
        "below_electrolyte",
        ConfigValue(
            domain=Bool,
            description="Whether the triple phase boundary is located below or "
            "above the electrolyte. This flag determines the sign of material_flux_x.",
        ),
    )
    CONFIG.declare(
        "voltage_drop_custom",
        ConfigValue(
            domain=Bool,
            default=False,
            description="If True, add voltage_drop_custom Var to be connected to degradation models",
        ),
    )

    common._submodel_boilerplate_config(CONFIG)
    common._thermal_boundary_conditions_config(CONFIG, thin=True)
    common._material_boundary_conditions_config(CONFIG, thin=True)

    def build(self):
        super().build()

        # Set up some sets for the space and time indexing
        tset = self.flowsheet().config.time
        # Set up node and face sets and get integer indices for them
        izfaces, iznodes = common._face_initializer(  # pylint: disable=unused-variable
            self, self.config.control_volume_zfaces, "z"
        )
        comps = self.component_list = pyo.Set(
            initialize=self.config.component_list,
            ordered=True,
            doc="Set of all gas-phase components present in submodel",
        )

        if "e^-" not in self.config.reaction_stoichiometry.keys():
            raise ConfigurationError(
                f"Number of electrons produced or consumed in redox reaction at {self.name} "
                "not specified."
            )

        self.reaction_stoichiometry = copy.copy(self.config.reaction_stoichiometry)

        if self.config.inert_species is None:
            self.config.inert_species = list()
        # Copy and pasted from the Gibbs reactor
        for j in self.config.inert_species:
            if j not in comps:
                raise ConfigurationError(
                    "{} invalid component in inert_species argument. {} is "
                    "not in the provided component list.".format(self.name, j)
                )

        self.inert_species_list = pyo.Set(
            initialize=self.config.inert_species,
            ordered=True,
            doc="Set of components that do not react at triple phase boundary",
        )
        # Ensure all inerts have been assigned a zero for a stoichiometric coefficient
        for j in self.inert_species_list:
            try:
                # Want to future-proof this method in case floating-point round-off ever becomes an issue.
                if abs(self.reaction_stoichiometry[j]) > 1e-8:
                    raise ConfigurationError(
                        f"Component {j} was in inert_species_list provided to {self.name}, but "
                        "has a nonzero stoichiometric coefficient."
                    )
            except KeyError:
                # Inert species does not have stoichiometry specified.
                pass
            self.reaction_stoichiometry[j] = 0

        self.reacting_component_list = pyo.Set(
            initialize=[
                j
                for j, coeff in self.reaction_stoichiometry.items()
                if j not in self.inert_species_list
            ],
            ordered=True,
            doc="Set of components (gas-phase and solid) that react at triple "
            "phase boundary",
        )
        self.reacting_gas_list = pyo.Set(
            initialize=[j for j in comps if j not in self.inert_species_list],
            ordered=True,
            doc="Set of gas-phase components that react at triple phase boundary",
        )

        common._submodel_boilerplate_create_if_none(self)
        common._create_thermal_boundary_conditions_if_none(self, thin=True)
        common._create_material_boundary_conditions_if_none(self, thin=True)

        common._create_if_none(
            self,
            "conc_mol_comp_ref",
            idx_set=(tset, iznodes, comps),
            units=pyo.units.mol / pyo.units.m**3,
        )

        self.mole_frac_comp = pyo.Var(
            tset,
            iznodes,
            comps,
            initialize=1 / len(comps),
            units=pyo.units.dimensionless,
            bounds=(0, 1),
        )

        self.log_mole_frac_comp = pyo.Var(
            tset,
            iznodes,
            self.reacting_gas_list,
            initialize=-1,
            units=pyo.units.dimensionless,
            bounds=(None, 0),
        )

        self.activation_potential = pyo.Var(
            tset,
            iznodes,
            initialize=1,
            units=pyo.units.V,
        )

        self.activation_potential_alpha1 = pyo.Var(
            initialize=0.5,
            units=pyo.units.dimensionless,
        )

        self.activation_potential_alpha2 = pyo.Var(
            initialize=0.5,
            units=pyo.units.dimensionless,
        )

        self.exchange_current_exponent_comp = pyo.Var(
            self.reacting_gas_list,
            initialize=1,
            units=pyo.units.dimensionless,
            bounds=(0, None),
        )

        # Nominally dimensionless, but specifically the log of the preexponential factor in amps per m**2
        self.exchange_current_log_preexponential_factor = pyo.Var(
            initialize=1, units=pyo.units.dimensionless, bounds=(0, None)
        )

        self.exchange_current_activation_energy = pyo.Var(
            initialize=0, units=pyo.units.J / pyo.units.mol, bounds=(0, None)
        )
        if self.config.voltage_drop_custom:
            self.voltage_drop_custom = pyo.Var(
                tset,
                iznodes,
                units=pyo.units.volts,
                doc="Custom voltage drop term for degradation modeling",
            )

        @self.Expression(tset, iznodes, comps)
        def conc_mol_comp(b, t, iz, j):
            return b.conc_mol_comp_ref[t, iz, j] + b.conc_mol_comp_deviation_x[t, iz, j]

        @self.Expression(tset, iznodes)
        def pressure(b, t, iz):
            return (
                sum(b.conc_mol_comp[t, iz, i] for i in comps)
                * Constants.gas_constant
                * b.temperature[t, iz]
            )

        # mole_frac_comp must be a variable because we want IPOPT to enforce
        # a lower bound of 0 in order to avoid AMPL errors, etc.
        @self.Constraint(tset, iznodes, comps)
        def mole_frac_comp_eqn(b, t, iz, j):
            return b.mole_frac_comp[t, iz, j] == b.conc_mol_comp[t, iz, j] / sum(
                b.conc_mol_comp[t, iz, i] for i in comps
            )

        @self.Constraint(tset, iznodes, self.reacting_gas_list)
        def log_mole_frac_comp_eqn(b, t, iz, j):
            return b.mole_frac_comp[t, iz, j] == pyo.exp(b.log_mole_frac_comp[t, iz, j])

        @self.Expression(tset, iznodes)
        def ds_rxn(b, t, iz):
            T = b.temperature[t, iz]
            P = b.pressure[t, iz]
            P_ref = 1e5 * pyo.units.Pa
            log_y_j = b.log_mole_frac_comp
            nu_j = b.reaction_stoichiometry
            # Any j not in comps is assumed to not be vapor phase
            pressure_exponent = sum(nu_j[j] for j in b.reacting_gas_list)
            if abs(pressure_exponent) < 1e-6:
                out_expr = 0
            else:
                out_expr = (
                    -Constants.gas_constant * pressure_exponent * pyo.log(P / P_ref)
                )
            return out_expr + (
                sum(
                    nu_j[j] * common._comp_entropy_expr(T, j)
                    for j in b.reacting_component_list
                )
                - Constants.gas_constant
                * sum(
                    nu_j[j] * log_y_j[t, iz, j]
                    for j in b.reacting_gas_list
                    # TODO verify that excluding solids is correct
                )
            )

        @self.Expression(tset, iznodes)
        def dh_rxn(b, t, iz):
            return sum(
                b.reaction_stoichiometry[j]
                * common._comp_enthalpy_expr(b.temperature[t, iz], j)
                for j in b.reacting_component_list
            )

        @self.Expression(tset, iznodes)
        def dg_rxn(b, t, iz):
            return b.dh_rxn[t, iz] - b.temperature[t, iz] * b.ds_rxn[t, iz]

        @self.Expression(tset, iznodes)
        def potential_nernst(b, t, iz):
            if b.config.below_electrolyte:
                return -b.dg_rxn[t, iz] / (
                    Constants.faraday_constant * b.reaction_stoichiometry["e^-"]
                )
            else:
                return -b.dg_rxn[t, iz] / (
                    Constants.faraday_constant * -b.reaction_stoichiometry["e^-"]
                )

        @self.Expression(tset, iznodes)
        def log_exchange_current_density(b, t, iz):
            T = b.temperature[t, iz]
            log_k = b.exchange_current_log_preexponential_factor[None]
            expo = b.exchange_current_exponent_comp
            E_A = b.exchange_current_activation_energy[None]
            out = log_k - E_A / (Constants.gas_constant * T)
            for j in b.reacting_gas_list:
                out += expo[j] * b.log_mole_frac_comp[t, iz, j]
            return out

        # Butler Volmer equation
        @self.Constraint(tset, iznodes)
        def activation_potential_eqn(b, t, iz):
            i = b.current_density[t, iz]
            log_i0 = b.log_exchange_current_density[t, iz]
            eta = b.activation_potential[t, iz]
            T = b.temperature[t, iz]
            alpha_1 = b.activation_potential_alpha1[None]
            alpha_2 = b.activation_potential_alpha2[None]
            exp_expr = Constants.faraday_constant * eta / (Constants.gas_constant * T)
            return i == pyo.units.A / pyo.units.m**2 * (
                pyo.exp(log_i0 + alpha_1 * exp_expr)
                - pyo.exp(log_i0 - alpha_2 * exp_expr)
            )

        @self.Expression(tset, iznodes)
        def reaction_rate_per_unit_area(b, t, iz):
            # Assuming there are no current leaks, the reaction rate can be
            # calculated directly from the current density
            if b.config.below_electrolyte:
                return b.current_density[t, iz] / (
                    Constants.faraday_constant * b.reaction_stoichiometry["e^-"]
                )
            else:
                return b.current_density[t, iz] / (
                    Constants.faraday_constant * -b.reaction_stoichiometry["e^-"]
                )

        # Put this expression in to prepare for a contact resistance term
        @self.Expression(tset, iznodes)
        def voltage_drop_total(b, t, iz):
            if self.config.voltage_drop_custom:
                return b.activation_potential[t, iz] + b.voltage_drop_custom[t, iz]
            else:
                return b.activation_potential[t, iz]

        @self.Constraint(tset, iznodes)
        def heat_flux_x_eqn(b, t, iz):
            return (
                b.heat_flux_x1[t, iz]
                == b.heat_flux_x0[t, iz]
                + b.current_density[t, iz]
                * b.voltage_drop_total[t, iz]  # Resistive heating
                - b.reaction_rate_per_unit_area[t, iz]  # Reversible heat of reaction
                * b.temperature[t, iz]
                * b.ds_rxn[t, iz]
            )

        @self.Constraint(tset, iznodes, comps)
        def material_flux_x_eqn(b, t, iz, j):
            if b.config.below_electrolyte:
                return (
                    b.material_flux_x[t, iz, j]
                    == -b.reaction_rate_per_unit_area[t, iz]
                    * b.reaction_stoichiometry[j]
                )
            else:
                return (
                    b.material_flux_x[t, iz, j]
                    == b.reaction_rate_per_unit_area[t, iz]
                    * b.reaction_stoichiometry[j]
                )

    def initialize_build(
        self, outlvl=idaeslog.NOTSET, solver=None, optarg=None, fix_x0=False
    ):
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        self.temperature_deviation_x.fix()
        self.conc_mol_comp_ref.fix()
        self.conc_mol_comp_deviation_x.fix()
        if fix_x0:
            self.heat_flux_x0.fix()
        else:
            self.heat_flux_x1.fix()

        for t in self.flowsheet().time:
            for iz in self.iznodes:
                denom = pyo.value(
                    sum(self.conc_mol_comp[t, iz, j] for j in self.component_list)
                )
                for j in self.component_list:
                    self.mole_frac_comp[t, iz, j].value = pyo.value(
                        self.conc_mol_comp[t, iz, j] / denom
                    )
                    if j in self.reacting_gas_list:
                        self.log_mole_frac_comp[t, iz, j].value = pyo.value(
                            pyo.log(self.mole_frac_comp[t, iz, j])
                        )

        solver_obj = get_solver(solver, optarg)
        common._init_solve_block(self, solver_obj, solve_log)

        self.temperature_deviation_x.unfix()
        self.conc_mol_comp_ref.unfix()
        self.conc_mol_comp_deviation_x.unfix()
        if fix_x0:
            self.heat_flux_x0.unfix()
        else:
            self.heat_flux_x1.unfix()

    def calculate_scaling_factors(self):
        pass

    def recursive_scaling(self):
        gsf = iscale.get_scaling_factor

        def ssf(c, s):
            iscale.set_scaling_factor(c, s, overwrite=False)

        sgsf = iscale.set_and_get_scaling_factor

        def cst(c, s):
            iscale.constraint_scaling_transform(c, s, overwrite=False)

        sy_def = 10  # Mole frac comp scaling

        for t in self.flowsheet().time:
            for iz in self.iznodes:
                ssf(self.activation_potential[t, iz], 10)
                if self.current_density[t, iz].is_reference():
                    si = gsf(self.current_density[t, iz].referent, default=1e-2)
                else:
                    si = gsf(self.current_density[t, iz], default=1e-2, warning=True)
                # TODO come back when I come up with a good formulation
                cst(self.activation_potential_eqn[t, iz], si)
                if self.heat_flux_x0[t, iz].is_reference():
                    gsf(self.heat_flux_x0[t, iz].referent, default=1e-2)
                else:
                    sqx0 = sgsf(self.heat_flux_x0[t, iz], 1e-2)
                if self.heat_flux_x1[t, iz].is_reference():
                    sqx1 = gsf(self.heat_flux_x1[t, iz].referent, 1e-2)
                else:
                    sqx1 = sgsf(self.heat_flux_x1[t, iz], 1e-2)
                sqx = min(sqx0, sqx1)
                cst(self.heat_flux_x_eqn[t, iz], sqx)

                for j in self.component_list:
                    # TODO Come back with good formulation for trace components
                    # and scale DConc and Cref
                    sy = sgsf(self.mole_frac_comp[t, iz, j], sy_def)
                    cst(self.mole_frac_comp_eqn[t, iz, j], sy)
                    smaterial_flux_x = sgsf(self.material_flux_x[t, iz, j], 1e-2)
                    cst(self.material_flux_x_eqn[t, iz, j], smaterial_flux_x)
                    if j in self.reacting_gas_list:
                        ssf(self.log_mole_frac_comp[t, iz, j], 1)
                        cst(self.log_mole_frac_comp_eqn[t, iz, j], sy)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for ControlVolumeBlockData, and for initializing the moving bed module

Author: Chinedu Okoli
"""

import pytest

from pyomo.environ import (
    ConcreteModel,
    check_optimal_termination,
    SolverStatus,
    value,
    Var,
    Constraint,
    TransformationFactory,
    Reference,
    units as pyo_units,
)
from pyomo.core.expr.visitor import identify_variables
from pyomo.dae.flatten import slice_component_along_sets
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.config import ConfigBlock
from pyomo.common.collections import ComponentSet
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
    unused_variables_set,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import (
    InitializationError,
    ConfigurationError,
)

# Import MBR unit model
from idaes.models_extra.gas_solid_contactors.unit_models.moving_bed import MBR

# Import property packages
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.gas_phase_thermo import (
    GasPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.solid_phase_thermo import (
    SolidPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.hetero_reactions import (
    HeteroReactionParameterBlock,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # Set up thermo props and reaction props
    m.fs.gas_properties = GasPhaseParameterBlock()
    m.fs.solid_properties = SolidPhaseParameterBlock()
    m.fs.hetero_reactions = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_properties,
        gas_property_package=m.fs.gas_properties,
    )

    m.fs.unit = MBR(
        gas_phase_config={"property_package": m.fs.gas_properties},
        solid_phase_config={
            "property_package": m.fs.solid_properties,
            "reaction_package": m.fs.hetero_reactions,
        },
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 17
    assert isinstance(m.fs.unit.config.gas_phase_config, ConfigBlock)
    assert isinstance(m.fs.unit.config.solid_phase_config, ConfigBlock)

    assert m.fs.unit.config.finite_elements == 10
    assert m.fs.unit.config.length_domain_set == [0.0, 1.0]
    assert m.fs.unit.config.transformation_method == "dae.finite_difference"
    assert m.fs.unit.config.transformation_scheme == "BACKWARD"
    assert m.fs.unit.config.collocation_points == 3
    assert m.fs.unit.config.flow_type == "counter_current"
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.componentTotal
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.enthalpyTotal
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.has_pressure_change is True

    # Check gas phase config arguments
    assert len(m.fs.unit.config.gas_phase_config) == 7
    assert m.fs.unit.config.gas_phase_config.has_equilibrium_reactions is False
    assert m.fs.unit.config.gas_phase_config.property_package is m.fs.gas_properties
    assert m.fs.unit.config.gas_phase_config.reaction_package is None

    # Check solid phase config arguments
    assert len(m.fs.unit.config.solid_phase_config) == 7
    assert m.fs.unit.config.solid_phase_config.has_equilibrium_reactions is False
    assert m.fs.unit.config.solid_phase_config.property_package is m.fs.solid_properties
    assert m.fs.unit.config.solid_phase_config.reaction_package is m.fs.hetero_reactions


# -----------------------------------------------------------------------------
class TestIronOC(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = MBR(
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # Fix geometry variables
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(128.20513)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(298.15)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(2.00e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.02499)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.00001)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.975)

        m.fs.unit.solid_inlet.flow_mass[0].fix(591.4)  # kg/s
        m.fs.unit.solid_inlet.particle_porosity[0].fix(0.27)  # (-)
        m.fs.unit.solid_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert hasattr(iron_oc.fs.unit, "gas_inlet")
        assert len(iron_oc.fs.unit.gas_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_inlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_inlet")
        assert len(iron_oc.fs.unit.solid_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_inlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.temperature, Var)

        assert hasattr(iron_oc.fs.unit, "gas_outlet")
        assert len(iron_oc.fs.unit.gas_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_outlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_outlet")
        assert len(iron_oc.fs.unit.solid_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_outlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.temperature, Var)

        assert isinstance(iron_oc.fs.unit.bed_area_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_phase_area, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_phase_area, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_super_vel, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_super_vel, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_phase_config_pressure_drop, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_solid_htc_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_phase_heat_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_phase_config_rxn_ext, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_comp_hetero_rxn, Constraint)

        assert number_variables(iron_oc) == 819
        assert number_total_constraints(iron_oc) == 781
        assert number_unused_variables(iron_oc) == 16

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.fixture(scope="class")
    def iron_oc_unscaled(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = MBR(
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # Fix geometry variables
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(128.20513)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(298.15)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(2.00e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.02499)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.00001)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.975)

        m.fs.unit.solid_inlet.flow_mass[0].fix(591.4)  # kg/s
        m.fs.unit.solid_inlet.particle_porosity[0].fix(0.27)  # (-)
        m.fs.unit.solid_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_unscaled(self, iron_oc_unscaled):
        initialization_tester(
            iron_oc_unscaled,
            optarg={"tol": 1e-6},
            gas_phase_state_args={
                "flow_mol": 128.20513,
                "temperature": 1183.15,
                "pressure": 2.00e5,
            },
            solid_phase_state_args={"flow_mass": 591.4, "temperature": 1183.15},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_unscaled(self, iron_oc_unscaled):
        results = solver.solve(iron_oc_unscaled)

        # Check for optimal solution
        assert check_optimal_termination(results)
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_scaling(self, iron_oc):

        # Set scaling gas phase for state variables
        iron_oc.fs.gas_properties.set_default_scaling("flow_mol", 1e-3)
        iron_oc.fs.gas_properties.set_default_scaling("pressure", 1e-5)
        iron_oc.fs.gas_properties.set_default_scaling("temperature", 1e-2)
        for comp in iron_oc.fs.gas_properties.component_list:
            iron_oc.fs.gas_properties.set_default_scaling(
                "mole_frac_comp", 1e1, index=comp
            )
        # Set scaling for gas phase thermophysical and transport properties
        iron_oc.fs.gas_properties.set_default_scaling("enth_mol", 1e-6)
        iron_oc.fs.gas_properties.set_default_scaling("enth_mol_comp", 1e-6)
        iron_oc.fs.gas_properties.set_default_scaling("cp_mol", 1e-6)
        iron_oc.fs.gas_properties.set_default_scaling("cp_mol_comp", 1e-6)
        iron_oc.fs.gas_properties.set_default_scaling("cp_mass", 1e-6)
        iron_oc.fs.gas_properties.set_default_scaling("entr_mol", 1e-2)
        iron_oc.fs.gas_properties.set_default_scaling("entr_mol_phase", 1e-2)
        iron_oc.fs.gas_properties.set_default_scaling("dens_mol", 1)
        iron_oc.fs.gas_properties.set_default_scaling("dens_mol_comp", 1)
        iron_oc.fs.gas_properties.set_default_scaling("dens_mass", 1e2)
        iron_oc.fs.gas_properties.set_default_scaling("visc_d_comp", 1e4)
        iron_oc.fs.gas_properties.set_default_scaling("diffusion_comp", 1e5)
        iron_oc.fs.gas_properties.set_default_scaling("therm_cond_comp", 1e2)
        iron_oc.fs.gas_properties.set_default_scaling("visc_d", 1e5)
        iron_oc.fs.gas_properties.set_default_scaling("therm_cond", 1e0)
        iron_oc.fs.gas_properties.set_default_scaling("mw", 1e2)

        # Set scaling for solid phase state variables
        iron_oc.fs.solid_properties.set_default_scaling("flow_mass", 1e-3)
        iron_oc.fs.solid_properties.set_default_scaling("particle_porosity", 1e2)
        iron_oc.fs.solid_properties.set_default_scaling("temperature", 1e-2)
        for comp in iron_oc.fs.solid_properties.component_list:
            iron_oc.fs.solid_properties.set_default_scaling(
                "mass_frac_comp", 1e1, index=comp
            )

        # Set scaling for solid phase thermophysical and transport properties
        iron_oc.fs.solid_properties.set_default_scaling("enth_mass", 1e-6)
        iron_oc.fs.solid_properties.set_default_scaling("enth_mol_comp", 1e-6)
        iron_oc.fs.solid_properties.set_default_scaling("cp_mol_comp", 1e-6)
        iron_oc.fs.solid_properties.set_default_scaling("cp_mass", 1e-6)
        iron_oc.fs.solid_properties.set_default_scaling("dens_mass_particle", 1e-2)
        iron_oc.fs.solid_properties.set_default_scaling("dens_mass_skeletal", 1e-2)

        MB = iron_oc.fs.unit  # alias to keep test lines short

        # Calculate scaling factors
        iscale.calculate_scaling_factors(MB)

        assert pytest.approx(0.01538, abs=1e-2) == iscale.get_scaling_factor(
            MB.bed_diameter
        )
        assert pytest.approx(0.003014, abs=1e-2) == iscale.get_scaling_factor(
            MB.bed_area
        )
        assert pytest.approx(0.003014, abs=1e-2) == iscale.get_scaling_factor(
            MB.gas_phase.area
        )
        assert pytest.approx(0.003014, abs=1e-2) == iscale.get_scaling_factor(
            MB.solid_phase.area
        )
        assert pytest.approx(0.068, abs=1e-2) == iscale.get_scaling_factor(
            MB.gas_solid_htc[0, 0]
        )
        assert pytest.approx(6666.67, abs=1e-2) == iscale.get_scaling_factor(
            MB.Re_particle[0, 0]
        )
        assert pytest.approx(0.099, abs=1e-2) == iscale.get_scaling_factor(
            MB.Pr_particle[0, 0]
        )
        assert pytest.approx(0.06858, abs=1e-2) == iscale.get_scaling_factor(
            MB.Nu_particle[0, 0]
        )
        for (t, x), v in MB.gas_phase.heat.items():
            assert pytest.approx(0.001, abs=1e-2) == iscale.get_scaling_factor(v)
        for (t, x), v in MB.solid_phase.heat.items():
            assert pytest.approx(0.001, abs=1e-2) == iscale.get_scaling_factor(v)

        for c in MB.bed_area_eqn.values():
            assert pytest.approx(
                0.0030, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.gas_phase_area.items():
            assert pytest.approx(
                0.0030, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.solid_phase_area.items():
            assert pytest.approx(
                0.0030, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.gas_super_vel.items():
            assert pytest.approx(
                0.001, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.solid_super_vel.items():
            assert pytest.approx(
                0.0001, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.gas_phase_config_pressure_drop.items():
            assert pytest.approx(
                0.0001, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, r), c in MB.solid_phase_config_rxn_ext.items():
            assert pytest.approx(
                3.0135e-5, abs=1e-4
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, p, j), c in MB.gas_comp_hetero_rxn.items():
            assert pytest.approx(
                0.01, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.solid_phase_heat_transfer.items():
            assert pytest.approx(
                1e-9, abs=1e-8
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.reynolds_number_particle.items():
            assert pytest.approx(
                6666, abs=100
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.prandtl_number.items():
            assert pytest.approx(
                0.1, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.nusselt_number_particle.items():
            assert pytest.approx(
                0.07, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.gas_solid_htc_eqn.items():
            assert pytest.approx(
                0.07, abs=1e-2
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in MB.gas_phase_heat_transfer.items():
            assert pytest.approx(
                1e-9, abs=1e-8
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        initialization_tester(
            iron_oc,
            optarg={"tol": 1e-6},
            gas_phase_state_args={
                "flow_mol": 128.20513,
                "temperature": 1183.15,
                "pressure": 2.00e5,
            },
            solid_phase_state_args={"flow_mass": 591.4, "temperature": 1183.15},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iron_oc):
        results = solver.solve(iron_oc)

        # Check for optimal solution
        assert check_optimal_termination(results)
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iron_oc):
        assert (
            pytest.approx(0.0479, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 0].value
        )
        assert (
            pytest.approx(0.5675, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 1].value
        )
        assert (
            pytest.approx(0.0039, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_solid[0].value
        )
        # Check the pressure drop that occurs across the bed
        assert (
            pytest.approx(198217.7068, abs=1e-2)
            == iron_oc.fs.unit.gas_outlet.pressure[0].value
        )
        assert (
            pytest.approx(1782.2932, abs=1e-2)
            == iron_oc.fs.unit.gas_inlet.pressure[0].value
            - iron_oc.fs.unit.gas_outlet.pressure[0].value
        )

    @pytest.mark.component
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iron_oc):
        # Conservation of material check
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_phase.properties[0, 0].mw,
            iron_oc.fs.unit.gas_phase.properties[0, 0].mw_eqn,
        )
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_phase.properties[0, 1].mw,
            iron_oc.fs.unit.gas_phase.properties[0, 1].mw_eqn,
        )
        mbal_gas = value(
            (
                iron_oc.fs.unit.gas_inlet.flow_mol[0]
                * iron_oc.fs.unit.gas_phase.properties[0, 0].mw
            )
            - (
                iron_oc.fs.unit.gas_outlet.flow_mol[0]
                * iron_oc.fs.unit.gas_phase.properties[0, 1].mw
            )
        )
        mbal_solid = value(
            iron_oc.fs.unit.solid_inlet.flow_mass[0]
            - iron_oc.fs.unit.solid_outlet.flow_mass[0]
        )
        mbal_tol = mbal_gas + mbal_solid
        assert abs(mbal_tol) <= 1e-2

        # Reaction stoichiometric ratio check
        # Overall reducer reactions for methane combustion:
        # CH4 + 12Fe2O3 => 8Fe3O4 + CO2 + 2H2O
        mole_gas_reacted = value(
            iron_oc.fs.unit.gas_inlet.flow_mol[0]
            * iron_oc.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"]
            - iron_oc.fs.unit.gas_outlet.flow_mol[0]
            * iron_oc.fs.unit.gas_outlet.mole_frac_comp[0, "CH4"]
        )
        mole_solid_reacted = value(
            (
                iron_oc.fs.unit.solid_inlet.flow_mass[0]
                * iron_oc.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_phase.properties[0, 1]._params.mw_comp["Fe2O3"]
            )
            - (
                iron_oc.fs.unit.solid_outlet.flow_mass[0]
                * iron_oc.fs.unit.solid_outlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_phase.properties[0, 0]._params.mw_comp["Fe2O3"]
            )
        )
        stoichiometric_ratio = mole_solid_reacted / mole_gas_reacted
        assert pytest.approx(12, abs=1e-6) == stoichiometric_ratio

        # Conservation of energy check
        ebal_gas = value(
            (
                iron_oc.fs.unit.gas_inlet.flow_mol[0]
                * iron_oc.fs.unit.gas_phase.properties[0, 0].enth_mol
            )
            - (
                iron_oc.fs.unit.gas_outlet.flow_mol[0]
                * iron_oc.fs.unit.gas_phase.properties[0, 1].enth_mol
            )
        )
        ebal_solid = value(
            (
                iron_oc.fs.unit.solid_inlet.flow_mass[0]
                * iron_oc.fs.unit.solid_phase.properties[0, 1].enth_mass
            )
            - (
                iron_oc.fs.unit.solid_outlet.flow_mass[0]
                * iron_oc.fs.unit.solid_phase.properties[0, 0].enth_mass
            )
        )
        e_reaction = value(
            mole_gas_reacted
            * iron_oc.fs.unit.solid_phase.reactions[0, 1]._params.dh_rxn["R1"]
        )
        ebal_tol = ebal_gas + ebal_solid - e_reaction
        assert abs(ebal_tol) <= 1e-2

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        iron_oc.fs.unit.report()

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        iron_oc.fs.unit.gas_outlet.flow_mol[0].fix(1)

        with pytest.raises(InitializationError):
            iron_oc.fs.unit.initialize()


# -----------------------------------------------------------------------------
class TestIronOC_EnergyBalanceType(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = MBR(
            energy_balance_type=EnergyBalanceType.none,
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # Fix geometry variables
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(128.20513)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(2.00e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.02499)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.00001)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.975)

        m.fs.unit.solid_inlet.flow_mass[0].fix(591.4)  # kg/s
        m.fs.unit.solid_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.solid_inlet.particle_porosity[0].fix(0.27)  # (-)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert hasattr(iron_oc.fs.unit, "gas_inlet")
        assert len(iron_oc.fs.unit.gas_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_inlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_inlet")
        assert len(iron_oc.fs.unit.solid_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_inlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.temperature, Var)

        assert hasattr(iron_oc.fs.unit, "gas_outlet")
        assert len(iron_oc.fs.unit.gas_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_outlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_outlet")
        assert len(iron_oc.fs.unit.solid_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_outlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.temperature, Var)

        assert isinstance(iron_oc.fs.unit.isothermal_gas_phase, Constraint)
        assert isinstance(iron_oc.fs.unit.isothermal_solid_phase, Constraint)

        assert number_variables(iron_oc) == 588
        assert number_total_constraints(iron_oc) == 508
        assert number_unused_variables(iron_oc) == 59
        print(unused_variables_set(iron_oc))

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.fixture(scope="class")
    def iron_oc_unscaled(self, iron_oc):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = MBR(
            energy_balance_type=EnergyBalanceType.none,
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # Fix geometry variables
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(128.20513)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(2.00e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.02499)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.00001)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.975)

        m.fs.unit.solid_inlet.flow_mass[0].fix(591.4)  # kg/s
        m.fs.unit.solid_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.solid_inlet.particle_porosity[0].fix(0.27)  # (-)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_unscaled(self, iron_oc_unscaled):
        initialization_tester(
            iron_oc_unscaled,
            optarg={"tol": 1e-6},
            gas_phase_state_args={
                "flow_mol": 128.20513,
                "temperature": 1183.15,
                "pressure": 2.00e5,
            },
            solid_phase_state_args={"flow_mass": 591.4, "temperature": 1183.15},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_unscaled(self, iron_oc_unscaled):
        results = solver.solve(iron_oc_unscaled)

        # Check for optimal solution
        assert check_optimal_termination(results)
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        iscale.calculate_scaling_factors(iron_oc.fs.unit)
        initialization_tester(
            iron_oc,
            optarg={"tol": 1e-6},
            gas_phase_state_args={
                "flow_mol": 128.20513,
                "temperature": 1183.15,
                "pressure": 2.00e5,
            },
            solid_phase_state_args={"flow_mass": 591.4, "temperature": 1183.15},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_scaled(self, iron_oc):
        results = solver.solve(iron_oc)

        # Check for optimal solution
        assert check_optimal_termination(results)
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution_scaled(self, iron_oc):
        assert (
            pytest.approx(0.1900, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 0].value
        )
        assert (
            pytest.approx(0.5675, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 1].value
        )
        assert (
            pytest.approx(0.0039, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_solid[0].value
        )
        # Check the pressure drop that occurs across the bed
        assert (
            pytest.approx(198214.8255, abs=1e-2)
            == iron_oc.fs.unit.gas_outlet.pressure[0].value
        )
        assert (
            pytest.approx(1785.1745, abs=1e-2)
            == iron_oc.fs.unit.gas_inlet.pressure[0].value
            - iron_oc.fs.unit.gas_outlet.pressure[0].value
        )

    @pytest.mark.component
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iron_oc):
        # Conservation of material check
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_phase.properties[0, 0].mw,
            iron_oc.fs.unit.gas_phase.properties[0, 0].mw_eqn,
        )
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_phase.properties[0, 1].mw,
            iron_oc.fs.unit.gas_phase.properties[0, 1].mw_eqn,
        )
        mbal_gas = value(
            (
                iron_oc.fs.unit.gas_inlet.flow_mol[0]
                * iron_oc.fs.unit.gas_phase.properties[0, 0].mw
            )
            - (
                iron_oc.fs.unit.gas_outlet.flow_mol[0]
                * iron_oc.fs.unit.gas_phase.properties[0, 1].mw
            )
        )
        mbal_solid = value(
            iron_oc.fs.unit.solid_inlet.flow_mass[0]
            - iron_oc.fs.unit.solid_outlet.flow_mass[0]
        )
        mbal_tol = mbal_gas + mbal_solid
        assert abs(mbal_tol) <= 1e-2

        # Reaction stoichiometric ratio check
        # Overall reducer reactions for methane combustion:
        # CH4 + 12Fe2O3 => 8Fe3O4 + CO2 + 2H2O
        mole_gas_reacted = value(
            iron_oc.fs.unit.gas_inlet.flow_mol[0]
            * iron_oc.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"]
            - iron_oc.fs.unit.gas_outlet.flow_mol[0]
            * iron_oc.fs.unit.gas_outlet.mole_frac_comp[0, "CH4"]
        )
        mole_solid_reacted = value(
            (
                iron_oc.fs.unit.solid_inlet.flow_mass[0]
                * iron_oc.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_phase.properties[0, 1]._params.mw_comp["Fe2O3"]
            )
            - (
                iron_oc.fs.unit.solid_outlet.flow_mass[0]
                * iron_oc.fs.unit.solid_outlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_phase.properties[0, 0]._params.mw_comp["Fe2O3"]
            )
        )
        stoichiometric_ratio = mole_solid_reacted / mole_gas_reacted
        assert pytest.approx(12, abs=1e-6) == stoichiometric_ratio

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        iron_oc.fs.unit.report()

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        iron_oc.fs.unit.gas_outlet.flow_mol[0].fix(1)

        with pytest.raises(InitializationError):
            iron_oc.fs.unit.initialize()


# -----------------------------------------------------------------------------
class TestIronOC_TransformationMethod(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = MBR(
            transformation_method="dae.collocation",
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # Fix geometry variables
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(128.20513)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(2.00e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.02499)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.00001)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.975)

        m.fs.unit.solid_inlet.flow_mass[0].fix(591.4)  # kg/s
        m.fs.unit.solid_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.solid_inlet.particle_porosity[0].fix(0.27)  # (-)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert hasattr(iron_oc.fs.unit, "gas_inlet")
        assert len(iron_oc.fs.unit.gas_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_inlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_inlet")
        assert len(iron_oc.fs.unit.solid_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_inlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.temperature, Var)

        assert hasattr(iron_oc.fs.unit, "gas_outlet")
        assert len(iron_oc.fs.unit.gas_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_outlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_outlet")
        assert len(iron_oc.fs.unit.solid_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_outlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.temperature, Var)

        assert number_variables(iron_oc) == 2279
        assert number_total_constraints(iron_oc) == 2241
        assert number_unused_variables(iron_oc) == 16
        print(unused_variables_set(iron_oc))

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.fixture(scope="class")
    def iron_oc_unscaled(self, iron_oc):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = MBR(
            transformation_method="dae.collocation",
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # Fix geometry variables
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(128.20513)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(2.00e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.02499)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.00001)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.975)

        m.fs.unit.solid_inlet.flow_mass[0].fix(591.4)  # kg/s
        m.fs.unit.solid_inlet.temperature[0].fix(1183.15)  # K
        m.fs.unit.solid_inlet.particle_porosity[0].fix(0.27)  # (-)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_unscaled(self, iron_oc_unscaled):
        initialization_tester(
            iron_oc_unscaled,
            optarg={"tol": 1e-5},
            gas_phase_state_args={
                "flow_mol": 128.20513,
                "temperature": 1183.15,
                "pressure": 2.00e5,
            },
            solid_phase_state_args={"flow_mass": 591.4, "temperature": 1183.15},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_unscaled(self, iron_oc_unscaled):
        solver.options = {"tol": 1e-5}
        results = solver.solve(iron_oc_unscaled)

        # Check for optimal solution
        assert check_optimal_termination(results)
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        iscale.calculate_scaling_factors(iron_oc.fs.unit)
        initialization_tester(
            iron_oc,
            optarg={"tol": 1e-5},
            gas_phase_state_args={
                "flow_mol": 128.20513,
                "temperature": 1183.15,
                "pressure": 2.00e5,
            },
            solid_phase_state_args={"flow_mass": 591.4, "temperature": 1183.15},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_scaled(self, iron_oc):
        solver.options = {"tol": 1e-5}
        results = solver.solve(iron_oc)

        # Check for optimal solution
        assert check_optimal_termination(results)
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution_scaled(self, iron_oc):
        assert (
            pytest.approx(0.1900, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 0].value
        )
        assert (
            pytest.approx(0.5675, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 1].value
        )
        assert (
            pytest.approx(0.0039, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_solid[0].value
        )
        # Check the pressure drop that occurs across the bed
        assert (
            pytest.approx(198214.8255, abs=1e3)
            == iron_oc.fs.unit.gas_outlet.pressure[0].value
        )
        assert (
            pytest.approx(1785.1745, abs=1e2)
            == iron_oc.fs.unit.gas_inlet.pressure[0].value
            - iron_oc.fs.unit.gas_outlet.pressure[0].value
        )

    @pytest.mark.component
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iron_oc):
        # Conservation of material check
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_phase.properties[0, 0].mw,
            iron_oc.fs.unit.gas_phase.properties[0, 0].mw_eqn,
        )
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_phase.properties[0, 1].mw,
            iron_oc.fs.unit.gas_phase.properties[0, 1].mw_eqn,
        )
        mbal_gas = value(
            (
                iron_oc.fs.unit.gas_inlet.flow_mol[0]
                * iron_oc.fs.unit.gas_phase.properties[0, 0].mw
            )
            - (
                iron_oc.fs.unit.gas_outlet.flow_mol[0]
                * iron_oc.fs.unit.gas_phase.properties[0, 1].mw
            )
        )
        mbal_solid = value(
            iron_oc.fs.unit.solid_inlet.flow_mass[0]
            - iron_oc.fs.unit.solid_outlet.flow_mass[0]
        )
        mbal_tol = mbal_gas + mbal_solid
        assert abs(mbal_tol) <= 1e-2

        # Reaction stoichiometric ratio check
        # Overall reducer reactions for methane combustion:
        # CH4 + 12Fe2O3 => 8Fe3O4 + CO2 + 2H2O
        mole_gas_reacted = value(
            iron_oc.fs.unit.gas_inlet.flow_mol[0]
            * iron_oc.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"]
            - iron_oc.fs.unit.gas_outlet.flow_mol[0]
            * iron_oc.fs.unit.gas_outlet.mole_frac_comp[0, "CH4"]
        )
        mole_solid_reacted = value(
            (
                iron_oc.fs.unit.solid_inlet.flow_mass[0]
                * iron_oc.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_phase.properties[0, 1]._params.mw_comp["Fe2O3"]
            )
            - (
                iron_oc.fs.unit.solid_outlet.flow_mass[0]
                * iron_oc.fs.unit.solid_outlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_phase.properties[0, 0]._params.mw_comp["Fe2O3"]
            )
        )
        stoichiometric_ratio = mole_solid_reacted / mole_gas_reacted
        assert pytest.approx(12, abs=1e-6) == stoichiometric_ratio

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        iron_oc.fs.unit.report()

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        iron_oc.fs.unit.gas_outlet.flow_mol[0].fix(1)

        with pytest.raises(InitializationError):
            iron_oc.fs.unit.initialize()


class TestBidirectionalSpatialDiscretization:
    @pytest.mark.unit
    def test_config_errors(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()

        with pytest.raises(
            ConfigurationError,
            match="transformation_method must be",
        ):
            m.fs.unit1 = MBR(
                transformation_method="dae.collocation",
                gas_transformation_scheme="BACKWARD",
                gas_phase_config={"property_package": m.fs.gas_properties},
                solid_phase_config={"property_package": m.fs.solid_properties},
            )
            m.fs.unit1.build()

        with pytest.raises(
            ConfigurationError,
            match=r"Either both.*must be set",
        ):
            m.fs.unit2 = MBR(
                gas_transformation_scheme="BACKWARD",
                gas_phase_config={"property_package": m.fs.gas_properties},
                solid_phase_config={"property_package": m.fs.solid_properties},
            )
            m.fs.unit2.build()

        with pytest.raises(
            ConfigurationError,
            match="transformation_scheme cannot be specified",
        ):
            m.fs.unit3 = MBR(
                transformation_scheme="BACKWARD",
                gas_transformation_scheme="BACKWARD",
                solid_transformation_scheme="BACKWARD",
                gas_phase_config={"property_package": m.fs.gas_properties},
                solid_phase_config={"property_package": m.fs.solid_properties},
            )
            m.fs.unit3.build()

    @pytest.mark.unit
    def test_construct_dynamic(self):
        horizon = 300.0
        tfe_width = 100.0
        ntcp = 3

        nxfe = 3
        nxcp = 1

        time_set = [0.0, horizon]
        ntfe = round(horizon / tfe_width)

        model = ConcreteModel()
        model.fs = FlowsheetBlock(
            dynamic=True,
            time_set=time_set,
            time_units=pyo_units.s,
        )

        model.fs.gas_properties = GasPhaseParameterBlock()
        model.fs.solid_properties = SolidPhaseParameterBlock()
        model.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=model.fs.solid_properties,
            gas_property_package=model.fs.gas_properties,
        )

        model.fs.MB = MBR(
            finite_elements=nxfe,
            has_holdup=True,
            length_domain_set=[],
            transformation_method="dae.finite_difference",
            collocation_points=nxcp,
            gas_transformation_scheme="BACKWARD",
            solid_transformation_scheme="FORWARD",
            pressure_drop_type="ergun_correlation",
            gas_phase_config={"property_package": model.fs.gas_properties},
            solid_phase_config={
                "property_package": model.fs.solid_properties,
                "reaction_package": model.fs.hetero_reactions,
            },
        )

        time = model.fs.time
        discretizer = TransformationFactory("dae.collocation")
        discretizer.apply_to(
            model, wrt=time, nfe=ntfe, ncp=ntcp, scheme="LAGRANGE-RADAU"
        )

        gas_phase = model.fs.MB.gas_phase
        solid_phase = model.fs.MB.solid_phase

        gas_length = model.fs.MB.gas_phase.length_domain
        solid_length = model.fs.MB.solid_phase.length_domain
        bed_length = model.fs.MB.length_domain

        gas_components = model.fs.gas_properties.component_list
        solid_components = model.fs.solid_properties.component_list
        gas_phases = model.fs.gas_properties.phase_list
        solid_phases = model.fs.solid_properties.phase_list

        assert gas_length is not solid_length
        assert len(gas_length) == len(solid_length)
        assert len(gas_length) == len(bed_length)
        for x in gas_length:
            assert x in solid_length
            assert x in bed_length

    @pytest.mark.unit
    def test_var_indices(self):
        horizon = 300.0
        tfe_width = 100.0
        ntcp = 3

        nxfe = 3
        nxcp = 1

        time_set = [0.0, horizon]
        ntfe = round(horizon / tfe_width)

        model = ConcreteModel()
        model.fs = FlowsheetBlock(
            dynamic=True,
            time_set=time_set,
            time_units=pyo_units.s,
        )

        model.fs.gas_properties = GasPhaseParameterBlock()
        model.fs.solid_properties = SolidPhaseParameterBlock()
        model.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=model.fs.solid_properties,
            gas_property_package=model.fs.gas_properties,
        )

        model.fs.MB = MBR(
            finite_elements=nxfe,
            has_holdup=True,
            length_domain_set=[],
            transformation_method="dae.finite_difference",
            collocation_points=nxcp,
            gas_transformation_scheme="BACKWARD",
            solid_transformation_scheme="FORWARD",
            pressure_drop_type="ergun_correlation",
            gas_phase_config={"property_package": model.fs.gas_properties},
            solid_phase_config={
                "property_package": model.fs.solid_properties,
                "reaction_package": model.fs.hetero_reactions,
            },
        )

        time = model.fs.time
        discretizer = TransformationFactory("dae.collocation")
        discretizer.apply_to(
            model, wrt=time, nfe=ntfe, ncp=ntcp, scheme="LAGRANGE-RADAU"
        )

        gas_phase = model.fs.MB.gas_phase
        solid_phase = model.fs.MB.solid_phase

        gas_length = model.fs.MB.gas_phase.length_domain
        solid_length = model.fs.MB.solid_phase.length_domain
        bed_length = model.fs.MB.length_domain

        gas_components = model.fs.gas_properties.component_list
        solid_components = model.fs.solid_properties.component_list
        gas_phases = model.fs.gas_properties.phase_list
        solid_phases = model.fs.solid_properties.phase_list

        gas_disc_eqs = [
            gas_phase.material_flow_dx_disc_eq,
            gas_phase.enthalpy_flow_dx_disc_eq,
        ]
        gas_flow_vars = [gas_phase._flow_terms, gas_phase._enthalpy_flow]

        gas_disc_eqs = [
            dict(
                (idx, Reference(slice_))
                for idx, slice_ in slice_component_along_sets(
                    eq, (gas_length, solid_length)
                )
            )
            for eq in gas_disc_eqs
        ]
        gas_flow_vars = [
            dict(
                (idx, Reference(slice_))
                for idx, slice_ in slice_component_along_sets(
                    var, (gas_length, solid_length)
                )
            )
            for var in gas_flow_vars
        ]

        for eq_dict, var_dict in zip(gas_disc_eqs, gas_flow_vars):
            # NOTE: I am relying on fact that equations and variables
            # here have same non-space indices, and that they are in
            # the same order.
            for idx in eq_dict:
                for x in gas_length:
                    if x != gas_length.first():
                        x_prev = gas_length.prev(x)
                        eqdata = eq_dict[idx][x]
                        vardata = var_dict[idx][x]
                        vardata_prev = var_dict[idx][x_prev]
                        var_set = ComponentSet(identify_variables(eqdata.expr))
                        assert len(var_set) == 3
                        assert vardata in var_set
                        assert vardata_prev in var_set

        solid_disc_eqs = [
            solid_phase.material_flow_dx_disc_eq,
            solid_phase.enthalpy_flow_dx_disc_eq,
        ]
        solid_flow_vars = [
            solid_phase._flow_terms,
            solid_phase._enthalpy_flow,
        ]

        solid_disc_eqs = [
            dict(
                (idx, Reference(slice_))
                for idx, slice_ in slice_component_along_sets(
                    eq, (gas_length, solid_length)
                )
            )
            for eq in solid_disc_eqs
        ]
        solid_flow_vars = [
            dict(
                (idx, Reference(slice_))
                for idx, slice_ in slice_component_along_sets(
                    var, (gas_length, solid_length)
                )
            )
            for var in solid_flow_vars
        ]

        gas_disc_eqs = [
            gas_phase.material_flow_dx_disc_eq,
            gas_phase.enthalpy_flow_dx_disc_eq,
        ]
        gas_flow_vars = [
            gas_phase._flow_terms,
            gas_phase._enthalpy_flow,
        ]

        gas_disc_eqs = [
            dict(
                (idx, Reference(slice_))
                for idx, slice_ in slice_component_along_sets(
                    eq, (gas_length, solid_length)
                )
            )
            for eq in gas_disc_eqs
        ]
        gas_flow_vars = [
            dict(
                (idx, Reference(slice_))
                for idx, slice_ in slice_component_along_sets(
                    var, (gas_length, solid_length)
                )
            )
            for var in gas_flow_vars
        ]

        #
        # Test that discretization equations contain variables at the correct
        # indices
        #
        for eq_dict, var_dict in zip(gas_disc_eqs, gas_flow_vars):
            for idx in eq_dict:
                for x in gas_length:
                    if x != gas_length.first():
                        x_prev = gas_length.prev(x)
                        eqdata = eq_dict[idx][x]
                        vardata = var_dict[idx][x]
                        vardata_prev = var_dict[idx][x_prev]
                        var_set = ComponentSet(identify_variables(eqdata.expr))
                        assert len(var_set) == 3
                        assert vardata in var_set
                        assert vardata_prev in var_set

        solid_disc_eqs = [
            solid_phase.material_flow_dx_disc_eq,
            solid_phase.enthalpy_flow_dx_disc_eq,
        ]
        solid_flow_vars = [
            solid_phase._flow_terms,
            solid_phase._enthalpy_flow,
        ]

        solid_disc_eqs = [
            dict(
                (idx, Reference(slice_))
                for idx, slice_ in slice_component_along_sets(
                    eq, (gas_length, solid_length)
                )
            )
            for eq in solid_disc_eqs
        ]
        solid_flow_vars = [
            dict(
                (idx, Reference(slice_))
                for idx, slice_ in slice_component_along_sets(
                    var, (gas_length, solid_length)
                )
            )
            for var in solid_flow_vars
        ]

        #
        # Test that discretization equations contain variables at the correct
        # indices
        #
        for eq_dict, var_dict in zip(solid_disc_eqs, solid_flow_vars):
            for idx in eq_dict:
                for x in solid_length:
                    if x != solid_length.last():
                        x_prev = solid_length.next(x)
                        eqdata = eq_dict[idx][x]
                        vardata = var_dict[idx][x]
                        vardata_prev = var_dict[idx][x_prev]
                        var_set = ComponentSet(identify_variables(eqdata.expr))
                        assert len(var_set) == 3
                        assert vardata in var_set
                        assert vardata_prev in var_set



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
IDAES General Purpose Packed Solvent Column Model.
"""

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    Expression,
    Param,
    Reals,
    NonNegativeReals,
    Var,
    SolverStatus,
    TerminationCondition,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, Bool, In
from pyomo.util.calc_var_value import calculate_variable_from_constraint

# Import IDAES Libraries
from idaes.core.util.constants import Constants
from idaes.core import (
    ControlVolume1DBlock,
    UnitModelBlockData,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    FlowDirection,
    useDefault,
    DistributedVars,
)
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.logger as idaeslog
from idaes.core.util.config import DefaultBool


__author__ = "Paul Akula, John Eslick, Anuja Deshpande, Andrew Lee"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("PackedColumn")
class PackedColumnData(UnitModelBlockData):
    """
    Packed Column Model Class.
    """

    CONFIG = ConfigBlock()
    # For now, only support steady-state
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
**default** = False. Solvent Columns do not yet support dynamic behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=useDefault,
            domain=DefaultBool,
            description="Holdup construction flag",
            doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - construct holdup terms,
**False** - do not construct holdup terms}""",
        ),
    )

    # Configuration template for phase specific  arguments
    _PhaseCONFIG = ConfigBlock()

    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=20,
            domain=int,
            description="Number of finite elements length domain",
            doc="""Number of finite elements to use when discretizing length
domain (default=20)""",
        ),
    )

    CONFIG.declare(
        "length_domain_set",
        ConfigValue(
            default=[0.0, 1.0],
            domain=list,
            description="List of points in length domain",
            doc="""length_domain_set - (optional) list of point to use to
initialize a new ContinuousSet if length_domain is not
provided (default = [0.0, 1.0])""",
        ),
    )

    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed, **default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )

    # Populate the phase side template to default values
    _PhaseCONFIG.declare(
        "property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a ParameterBlock object""",
        ),
    )

    _PhaseCONFIG.declare(
        "property_package_args",
        ConfigValue(
            default={},
            description="Arguments for constructing vapor property package",
            doc="""A dict of arguments to be passed to the PropertyBlockData
and used when constructing these
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a dict (see property package for documentation)
            """,
        ),
    )

    # Create individual config blocks for vapor(gas) and liquid sides
    CONFIG.declare("vapor_phase", _PhaseCONFIG(doc="vapor side config arguments"))

    CONFIG.declare("liquid_phase", _PhaseCONFIG(doc="liquid side config arguments"))

    # =========================================================================

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to build default attributes
        super().build()

        # Check phase lists match assumptions
        if self.config.vapor_phase.property_package.phase_list != ["Vap"]:
            raise ConfigurationError(
                f"{self.name} SolventCondenser model requires that the vapor "
                f"phase property package have a single phase named 'Vap'"
            )
        if self.config.liquid_phase.property_package.phase_list != ["Liq"]:
            raise ConfigurationError(
                f"{self.name} SolventCondenser model requires that the liquid "
                f"phase property package have a single phase named 'Liq'"
            )

        # Check for at least one common component in component lists
        if not any(
            j in self.config.vapor_phase.property_package.component_list
            for j in self.config.liquid_phase.property_package.component_list
        ):
            raise ConfigurationError(
                f"{self.name} SolventCondenser model requires that the liquid "
                f"and vapor phase property packages have at least one "
                f"common component."
            )

        # Geometry
        self.diameter_column = Var(
            domain=Reals, initialize=0.1, units=pyunits.m, doc="Column diameter"
        )

        self.area_column = Var(
            domain=Reals,
            initialize=0.5,
            units=pyunits.m**2,
            doc="Column cross-sectional area",
        )

        self.length_column = Var(
            domain=Reals, initialize=4.9, units=pyunits.m, doc="Column length"
        )

        @self.Constraint(doc="Column cross-sectional area")
        def column_cross_section_area_eqn(blk):
            return blk.area_column == (Constants.pi * 0.25 * (blk.diameter_column) ** 2)

        # =====================================================================
        # Set argument values for vapor and liquid sides

        # Set flow directions for the control volume blocks
        # Gas flows from 0 to 1, Liquid flows from 1 to 0

        # TODO : Add support for different discretization schemes in future
        # TODO : Add support for dynamics in future (if required)
        # TODO: Only handling countercurrent flow for now.
        set_direction_vapor = FlowDirection.forward
        set_direction_liquid = FlowDirection.backward

        # =====================================================================
        # Build Control volume 1D for vapor phase and populate vapor control volume

        self.vapor_phase = ControlVolume1DBlock(
            transformation_method="dae.finite_difference",
            transformation_scheme="BACKWARD",
            finite_elements=self.config.finite_elements,
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            area_definition=DistributedVars.variant,
            property_package=self.config.vapor_phase.property_package,
            property_package_args=self.config.vapor_phase.property_package_args,
        )

        self.vapor_phase.add_geometry(
            flow_direction=set_direction_vapor,
            length_domain_set=self.config.length_domain_set,
            length_var=self.length_column,
        )

        self.vapor_phase.add_state_blocks(
            information_flow=set_direction_vapor, has_phase_equilibrium=False
        )

        self.vapor_phase.add_material_balances(
            balance_type=MaterialBalanceType.componentTotal,
            has_phase_equilibrium=False,
            has_mass_transfer=True,
        )

        self.vapor_phase.add_energy_balances(
            balance_type=EnergyBalanceType.enthalpyTotal,
            has_heat_transfer=True,
            has_enthalpy_transfer=True,
        )

        self.vapor_phase.add_momentum_balances(
            balance_type=MomentumBalanceType.pressureTotal,
            has_pressure_change=self.config.has_pressure_change,
        )

        self.vapor_phase.apply_transformation()

        # =====================================================================
        # Build Control volume 1D for liquid phase and populate liquid control volume

        self.liquid_phase = ControlVolume1DBlock(
            transformation_method="dae.finite_difference",
            transformation_scheme="FORWARD",
            finite_elements=self.config.finite_elements,
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            area_definition=DistributedVars.variant,
            property_package=self.config.liquid_phase.property_package,
            property_package_args=self.config.liquid_phase.property_package_args,
        )

        self.liquid_phase.add_geometry(
            flow_direction=set_direction_liquid,
            length_domain_set=self.config.length_domain_set,
            length_var=self.length_column,
        )

        self.liquid_phase.add_state_blocks(
            information_flow=set_direction_liquid, has_phase_equilibrium=False
        )

        self.liquid_phase.add_material_balances(
            balance_type=MaterialBalanceType.componentTotal,
            has_phase_equilibrium=False,
            has_mass_transfer=True,
        )

        self.liquid_phase.add_energy_balances(
            balance_type=EnergyBalanceType.enthalpyTotal,
            has_heat_transfer=True,
            has_enthalpy_transfer=True,
        )

        self.liquid_phase.apply_transformation()

        # Add Ports for vapor side
        self.add_inlet_port(name="vapor_inlet", block=self.vapor_phase)
        self.add_outlet_port(name="vapor_outlet", block=self.vapor_phase)

        # Add Ports for liquid side
        self.add_inlet_port(name="liquid_inlet", block=self.liquid_phase)
        self.add_outlet_port(name="liquid_outlet", block=self.liquid_phase)

        # TODO : Fix units

        # ======================================================================
        # Unit level sets
        vap_comp = self.config.vapor_phase.property_package.component_list
        liq_comp = self.config.liquid_phase.property_package.component_list
        equilibrium_comp = vap_comp & liq_comp

        # Get units of measurement - liquid phase units will be used as basis
        t_init = self.flowsheet().time.first()
        if (
            self.liquid_phase.properties[t_init, 0].get_material_flow_basis()
            != self.vapor_phase.properties[t_init, 0].get_material_flow_basis()
        ):
            raise ConfigurationError(
                f"{self.name} vapor and liquid property packages must use the "
                f"same material flow basis."
            )

        lunits = (
            self.config.liquid_phase.property_package.get_metadata().get_derived_units
        )

        # Hydrodynamics and cacking parameters
        self.eps_ref = Param(
            initialize=0.97,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Packing void space m3/m3",
        )

        self.packing_specific_area = Param(
            initialize=250,
            units=lunits("length") ** 2 / lunits("length") ** 3,
            mutable=True,
            doc="Packing specific surface area",
        )

        self.packing_channel_size = Param(
            initialize=0.1,
            units=lunits("length"),
            mutable=True,
            doc="Packing channel size",
        )

        self.hydraulic_diameter = Expression(
            expr=4 * self.eps_ref / self.packing_specific_area, doc="Hydraulic diameter"
        )

        self.area_interfacial = Var(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            initialize=0.9,
            units=(pyunits.m) ** 2 / (pyunits.m) ** 3,
            doc="Specific interfacial area",
        )

        # Liquid and vapor holdups
        self.holdup_liq = Var(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            initialize=0.001,
            units=pyunits.dimensionless,
            doc="Volumetric liquid holdup [-]",
        )

        # TODO : Consider making this a Var & Constraint
        def rule_holdup_vap(blk, t, x):
            if x == self.vapor_phase.length_domain.first():
                return Expression.Skip
            else:
                zb = self.vapor_phase.length_domain.prev(x)
                return blk.eps_ref - blk.holdup_liq[t, zb]

        self.holdup_vap = Expression(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            rule=rule_holdup_vap,
            doc="Volumetric vapor holdup [-]",
        )

        # Area of control volume : vapor side and liquid side
        @self.Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            doc="Vapor phase cross-sectional area constraint",
        )
        def vapor_phase_area(blk, t, x):
            if x == self.vapor_phase.length_domain.first():
                return blk.vapor_phase.area[t, x] == (blk.eps_ref * blk.area_column)
            else:
                return blk.vapor_phase.area[t, x] == (
                    blk.area_column * blk.holdup_vap[t, x]
                )

        @self.Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            doc="Liquid phase cross-sectional area constraint",
        )
        def liquid_phase_area(blk, t, x):
            if x == self.liquid_phase.length_domain.last():
                return blk.liquid_phase.area[t, x] == (blk.eps_ref * blk.area_column)
            else:
                return blk.liquid_phase.area[t, x] == (
                    blk.area_column * blk.holdup_liq[t, x]
                )

        # =====================================================================
        # Add performance equations
        # Pressure equality in phases
        @self.Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            doc="Mechanical equilibrium constraint",
        )
        def mechanical_equilibrium(blk, t, x):
            if x == self.liquid_phase.length_domain.last():
                return Constraint.Skip
            else:
                return blk.liquid_phase.properties[t, x].pressure == pyunits.convert(
                    blk.vapor_phase.properties[t, x].pressure,
                    to_units=lunits("pressure"),
                )

        # ---------------------------------------------------------------------
        # Mass transfer relationships
        self.mass_transfer_coeff_vap = Var(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            equilibrium_comp,
            units=(
                lunits("amount")
                / lunits("pressure")
                / lunits("length") ** 2
                / lunits("time")
            ),
            doc="Vapor phase mass transfer coefficient",
        )

        # Equilibrium partial pressure of components at interface
        self.pressure_equil = Var(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            equilibrium_comp,
            domain=NonNegativeReals,
            initialize=500,
            units=lunits("pressure"),
            doc="Equilibrium pressure of components at interface",
        )

        # Mass transfer constraints
        self.interphase_mass_transfer = Var(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            equilibrium_comp,
            domain=Reals,
            initialize=0.1,
            units=lunits("amount") / (lunits("time") * lunits("length")),
            doc="Interphase mass transfer rate",
        )

        self.liquid_phase_mass_transfer_model()

        @self.Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            equilibrium_comp,
            doc="Interphase mass transfer constraint",
        )
        def interphase_mass_transfer_eqn(blk, t, x, j):
            if x == self.vapor_phase.length_domain.first():
                return Constraint.Skip
            elif j in equilibrium_comp:
                return blk.interphase_mass_transfer[t, x, j] == (
                    blk.mass_transfer_coeff_vap[t, x, j]
                    * blk.area_interfacial[t, x]
                    * blk.area_column
                    * (
                        blk.vapor_phase.properties[t, x].mole_frac_comp[j]
                        * pyunits.convert(
                            blk.vapor_phase.properties[t, x].pressure,
                            to_units=lunits("pressure"),
                        )
                        - blk.pressure_equil[t, x, j]
                    )
                )
            else:
                return blk.interphase_mass_transfer[t, x, j] == 0.0

        @self.Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            self.liquid_phase.properties.component_list,
            doc="Liquid phase mass transfer constraint",
        )
        def liquid_mass_transfer_eqn(blk, t, x, j):
            if x == self.liquid_phase.length_domain.last():
                return Constraint.Skip
            elif j in equilibrium_comp:
                zf = self.vapor_phase.length_domain.next(x)
                return (
                    blk.liquid_phase.mass_transfer_term[t, x, "Liq", j]
                    == blk.interphase_mass_transfer[t, zf, j]
                )
            else:
                return blk.liquid_phase.mass_transfer_term[t, x, "Liq", j] == 0.0

        @self.Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            self.vapor_phase.properties.component_list,
            doc="Vapor phase mass transfer constraint",
        )
        def vapor_mass_transfer_eqn(blk, t, x, j):
            if x == self.vapor_phase.length_domain.first():
                return Constraint.Skip
            elif j in equilibrium_comp:
                return (
                    pyunits.convert(
                        blk.vapor_phase.mass_transfer_term[t, x, "Vap", j],
                        to_units=lunits("amount") / lunits("time") / lunits("length"),
                    )
                    == -blk.interphase_mass_transfer[t, x, j]
                )
            else:
                return blk.vapor_phase.mass_transfer_term[t, x, "Vap", j] == 0.0

        # ---------------------------------------------------------------------
        # Heat transfer relationships
        self.heat_transfer_coeff = Var(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            initialize=100,
            units=lunits("power") / lunits("temperature") / lunits("length"),
            doc="Vapor-liquid heat transfer coefficient",
        )

        # Heat transfer
        @self.Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            doc="Heat transfer calculation",
        )
        def heat_transfer_eqn1(blk, t, x):
            if x == self.vapor_phase.length_domain.first():
                return Constraint.Skip
            else:
                zb = self.liquid_phase.length_domain.prev(x)
                return pyunits.convert(
                    blk.vapor_phase.heat[t, x],
                    to_units=lunits("power") / lunits("length"),
                ) == (
                    blk.heat_transfer_coeff[t, x]
                    * (
                        blk.liquid_phase.properties[t, zb].temperature
                        - pyunits.convert(
                            blk.vapor_phase.properties[t, x].temperature,
                            to_units=lunits("temperature"),
                        )
                    )
                )

        @self.Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            doc="Heat transfer balance",
        )
        def heat_transfer_eqn2(blk, t, x):
            if x == self.liquid_phase.length_domain.last():
                return Constraint.Skip
            else:
                zf = self.vapor_phase.length_domain.next(x)
                return blk.liquid_phase.heat[t, x] == -pyunits.convert(
                    blk.vapor_phase.heat[t, zf],
                    to_units=lunits("power") / lunits("length"),
                )

        @self.Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            doc="Enthalpy transfer calculation",
        )
        def enthalpy_transfer_eqn1(blk, t, x):
            if x == self.vapor_phase.length_domain.first():
                return Constraint.Skip
            else:
                return blk.vapor_phase.enthalpy_transfer[t, x] == (
                    (
                        sum(
                            blk.vapor_phase.properties[t, x].enth_mol_phase_comp[
                                "Vap", j
                            ]
                            * blk.vapor_phase.mass_transfer_term[t, x, "Vap", j]
                            for j in equilibrium_comp
                        )
                    )
                )

        @self.Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            doc="Enthalpy transfer equality",
        )
        def enthalpy_transfer_eqn2(blk, t, x):
            if x == self.liquid_phase.length_domain.last():
                return Constraint.Skip
            else:
                zf = self.vapor_phase.length_domain.next(x)
                return blk.liquid_phase.enthalpy_transfer[t, x] == -pyunits.convert(
                    blk.vapor_phase.enthalpy_transfer[t, zf],
                    to_units=lunits("power") / lunits("length"),
                )

    def liquid_phase_mass_transfer_model(self):
        """
        Liquid phase mass transfer sub-model for calculating equilbrium
        partial pressure used in driving force.

        For the generic model, it is assumed that liquid phase mass transfer
        is sufficiently fast that the equilibrium partial pressure is equal to
        the fugacity of each component in the bulk.

        Derived Classes should overload this method as required.
        """
        vap_comp = self.config.vapor_phase.property_package.component_list
        liq_comp = self.config.liquid_phase.property_package.component_list
        equilibrium_comp = vap_comp & liq_comp

        @self.Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            equilibrium_comp,
            doc="Equilibrium partial pressure at interface",
        )
        def pressure_at_interface(blk, t, x, j):
            if x == self.vapor_phase.length_domain.first():
                return Constraint.Skip
            else:
                zb = self.liquid_phase.length_domain.prev(x)
                lprops = blk.liquid_phase.properties[t, zb]
                return blk.pressure_equil[t, x, j] == (lprops.fug_phase_comp["Liq", j])

    # =========================================================================
    # Scaling routine
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # ---------------------------------------------------------------------
        # Scale variables
        for (t, x, j), v in self.pressure_equil.items():
            if iscale.get_scaling_factor(v) is None:
                sf_pe = iscale.get_scaling_factor(
                    self.pressure_equil, default=None, warning=True
                )
                if sf_pe is None:
                    sf_pe = iscale.get_scaling_factor(
                        self.liquid_phase.properties[t, x].fug_phase_comp["Liq", j],
                        default=None,
                        warning=True,
                    )
                if sf_pe is None:
                    sf_pe = iscale.get_scaling_factor(
                        self.liquid_phase.properties[t, x].pressure,
                        default=1,
                        warning=True,
                    )

                iscale.set_scaling_factor(v, sf_pe)

        for (t, x), v in self.vapor_phase.heat.items():
            if iscale.get_scaling_factor(v) is None:
                sf = iscale.get_scaling_factor(
                    self.heat_transfer_coeff, default=1e-3, warning=True
                )
                iscale.set_scaling_factor(v, sf)

        for (t, x), v in self.liquid_phase.heat.items():
            if iscale.get_scaling_factor(v) is None:
                sf = iscale.get_scaling_factor(
                    self.heat_transfer_coeff, default=1e-3, warning=True
                )
                iscale.set_scaling_factor(v, sf)

        # ---------------------------------------------------------------------
        # Scale constraints
        for (t, x), v in self.mechanical_equilibrium.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    self.vapor_phase.properties[t, x].pressure, default=1, warning=True
                ),
            )

        for (t, x, j), v in self.pressure_at_interface.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    self.pressure_equil[t, x, j], default=1, warning=False
                ),
            )

        for (t, x, j), v in self.interphase_mass_transfer_eqn.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    self.interphase_mass_transfer[t, x, j], default=1, warning=True
                ),
            )

        for (t, x, j), v in self.liquid_mass_transfer_eqn.items():
            try:
                sf = iscale.get_scaling_factor(
                    self.interphase_mass_transfer[t, x, j], default=1, warning=False
                )
            except KeyError:
                # This implies a non-volatile component
                sf = 1
            iscale.constraint_scaling_transform(v, sf)

        for (t, x, j), v in self.vapor_mass_transfer_eqn.items():
            try:
                sf = iscale.get_scaling_factor(
                    self.interphase_mass_transfer[t, x, j], default=1, warning=False
                )
            except KeyError:
                # This implies a non-volatile component
                sf = 1
            iscale.constraint_scaling_transform(v, sf)

        for (t, x), v in self.heat_transfer_eqn1.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    self.vapor_phase.heat[t, x], default=1, warning=True
                ),
            )

        for (t, x), v in self.heat_transfer_eqn2.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    self.vapor_phase.heat[t, x], default=1, warning=True
                ),
            )

    # =========================================================================
    # Model initialization routine
    def initialize(
        blk,
        vapor_phase_state_args=None,
        liquid_phase_state_args=None,
        state_vars_fixed=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Standard Packed Column initialization.

        Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = None).
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during initialization
                    (default = None, use IDAES default solver)
        """

        # Set up logger for initialization and solve
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Set solver options
        opt = get_solver(solver, optarg)

        unit_constraints = [
            "pressure_at_interface",
            "interphase_mass_transfer_eqn",
            "liquid_mass_transfer_eqn",
            "vapor_mass_transfer_eqn",
            "heat_transfer_eqn1",
            "heat_transfer_eqn2",
            "enthalpy_transfer_eqn1",
            "enthalpy_transfer_eqn2",
        ]

        # ---------------------------------------------------------------------
        # Deactivate unit model level constraints
        for c in blk.component_objects(Constraint, descend_into=True):
            if c.local_name in unit_constraints:
                c.deactivate()

        # Fix variables

        # Interface pressure
        blk.pressure_equil.fix()

        # Molar flux
        blk.interphase_mass_transfer.fix(0.0)
        blk.vapor_phase.mass_transfer_term.fix(0.0)
        blk.liquid_phase.mass_transfer_term.fix(0.0)

        # Heat transfer rate
        blk.vapor_phase.heat.fix(0.0)
        blk.liquid_phase.heat.fix(0.0)
        blk.vapor_phase.enthalpy_transfer.fix(0.0)
        blk.liquid_phase.enthalpy_transfer.fix(0.0)

        # ---------------------------------------------------------------------
        # Provide state arguments for property package initialization

        init_log.info("Step 1: Property Package initialization")
        # Initialize vapor_phase properties block
        vflag = blk.vapor_phase.initialize(
            state_args=vapor_phase_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        # Initialize liquid_phase properties block
        lflag = blk.liquid_phase.initialize(
            state_args=liquid_phase_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        init_log.info("Step 2: Steady-State isothermal mass balance")

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Step 2: {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        init_log.info("Step 3: Interface equilibrium")

        # Activate interface pressure constraint
        blk.pressure_equil.unfix()
        blk.pressure_at_interface.activate()

        for k in blk.pressure_at_interface:
            calculate_variable_from_constraint(
                blk.pressure_equil[k], blk.pressure_at_interface[k]
            )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Step 3 complete: {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        init_log.info("Step 4a: Isothermal absoption")
        init_log.info_high("Calculating mass flux")

        # Unfix mass transfer terms
        blk.interphase_mass_transfer.unfix()

        # Activate mass transfer equation in vapor phase
        blk.interphase_mass_transfer_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        init_log.info("Step 4b: Isothermal chemical absoption")
        init_log.info_high("Adding mass transfer to material balances")

        blk.vapor_phase.mass_transfer_term.unfix()
        blk.liquid_phase.mass_transfer_term.unfix()
        blk.vapor_mass_transfer_eqn.activate()
        blk.liquid_mass_transfer_eqn.activate()

        # Fix this
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Step 4 complete: {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        init_log.info("Step 5: Adiabatic chemical absoption")
        init_log.info_high("Isothermal to Adiabatic ")

        # Unfix heat transfer terms
        blk.vapor_phase.heat.unfix()
        blk.liquid_phase.heat.unfix()
        blk.vapor_phase.enthalpy_transfer.unfix()
        blk.liquid_phase.enthalpy_transfer.unfix()

        # Activate heat transfer equations
        for c in [
            "heat_transfer_eqn1",
            "heat_transfer_eqn2",
            "enthalpy_transfer_eqn1",
            "enthalpy_transfer_eqn2",
        ]:
            getattr(blk, c).activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        init_log.info_high("Step 5 complete: {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        blk.vapor_phase.release_state(flags=vflag)
        blk.liquid_phase.release_state(flags=lflag)

        if (
            res.solver.termination_condition != TerminationCondition.optimal
            or res.solver.status != SolverStatus.ok
        ):
            raise InitializationError(
                f"{blk.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )
